from tw_experimentation.utils import ExperimentDataset, highlight
from tw_experimentation.statistical_tests import (
    FrequentistTest,
    FrequentistTestResults,
    compute_frequentist_results,
)
from typing import List, Optional, Dict, Union
from dataclasses import dataclass

import warnings
import copy

import pandas as pd
import json

from wise_pizza import explain_levels
from wise_pizza.plotting import plot_segments
import plotly.graph_objects as go
from IPython.display import display


@dataclass
class SegmentationOutput:
    """Represents the output of a segmentation analysis.

    Attributes:
        segments (List[str]): List of segment names.
        wise_pizza (pd.DataFrame): DataFrame containing the results of the
            analysis for each segment.
        segment_results (Dict[str, Dict[str, FrequentistTestResults]]):
            Dictionary containing the results of the frequentist tests for each segment.
            segment_results[category][segment] contains for each category the
                frequentist output for the segment.

    Methods:
        segment_output(segment: str) -> pd.DataFrame:
            Returns the results table for a specific segment.
    """

    segments: List[str]
    wise_pizza_figs: Dict[str, Dict[str, go.Figure]]
    wise_pizza_results: Dict[str, Dict[str, pd.DataFrame]]
    segment_results: Dict[str, Dict[str, FrequentistTestResults]]

    def segment_output(
        self, category: str, multitest_correction="bonferroni", alpha=0.05
    ) -> pd.DataFrame:
        """Returns the results table for a specific segment.

        Args:
            segment (str): The name of the segment.

        Returns:
            pd.DataFrame: The results table for the specified segment.
        """
        segment_tables = {
            k: self.segment_results[category][k]
            .compute_stats_per_target(
                multitest_correction=multitest_correction, type_i_error=alpha
            )
            .get_results_table()
            for k in self.segment_results[category]
        }
        results_table = pd.concat(
            segment_tables.values(), axis=0, keys=segment_tables.keys()
        )
        results_table.index.names = [category] + segment_tables[
            list(segment_tables.keys())[0]
        ].index.names[:]

        results_table = results_table.swaplevel(0, 2).sort_index()
        return results_table

    def wise_pizza_output(
        self,
        target: str,
        treatment: int,
        auto_display_df=False,
        multitest_correction="bonferroni",
    ):
        wise_pizza_df = {
            k: self.wise_pizza_results[target][treatment][k]
            .compute_stats_per_target(multitest_correction=multitest_correction)
            .get_results_table()
            for k in self.wise_pizza_results[target][treatment].keys()
        }
        results_table = pd.concat(
            wise_pizza_df.values(),
            axis=0,
            keys=wise_pizza_df.keys(),
        )
        results_table.index.names = ["Pizza slices", "Variants", "Outcomes"]
        results_table = results_table.swaplevel(0, 2)
        results_table = results_table.sort_index()
        if auto_display_df:
            display(
                results_table.style.apply(highlight, axis=1)
                .set_table_attributes("style='display:inline'")
                .bar(subset=["Estimated_Effect_relative"], color="grey")
                .format(precision=3)
            )
        return results_table


def run_segmentation_analysis(ed: ExperimentDataset, segments: Union[None, List[str]]):
    if segments is None:
        segments = ed.segments
    assert len(segments) > 0, "No segments specified"
    assert all([s in ed.data.columns for s in segments]), "Segment not in data"
    WP_MIN_SEGMENTS = 2
    WP_MAX_DEPTH = 3

    segmentation = SegmentationNew(ed)
    segment_outputs = {k: segmentation.frequentist_by_segment(k) for k in segments}
    wise_pizza_frequentist = dict()
    wise_pizza_figs = dict()
    for target in ed.targets:
        wise_pizza_figs[target] = dict()
        wise_pizza_frequentist[target] = dict()
        for treatment in range(1, ed.n_variants):
            (
                wise_pizza_frequentist[target][treatment],
                wise_pizza_figs[target][treatment],
            ) = segmentation.wise_pizza_frequentist(
                target=target,
                treatment=treatment,
                segments=segments,
                max_depth=WP_MAX_DEPTH,
                min_segments=WP_MIN_SEGMENTS,
            )

    so = SegmentationOutput(
        segments=segments,
        wise_pizza_figs=wise_pizza_figs,
        wise_pizza_results=wise_pizza_frequentist,
        segment_results=segment_outputs,
    )

    return so


class SegmentationNew:
    def __init__(self, ed) -> None:
        self.ed = ed
        self.df_wp = None
        self.sf = None

    def frequentist_by_segment(self, category: str, max_segments: int = 7):
        assert category in self.ed.data.columns

        segments = (
            self.ed.data.groupby(category)
            .count()
            .sort_values(by=[self.ed.variant], ascending=False)
            .index.values.tolist()[:max_segments]
        )
        segment_results = {}
        for segment in segments:
            ed_segment = copy.deepcopy(self.ed)
            ed_segment.data = ed_segment.data[ed_segment.data[category] == segment]
            segment_results[segment] = compute_frequentist_results(ed_segment)
        return segment_results

    def wise_pizza_frequentist(
        self,
        target: str,
        treatment: int = 1,
        segments=List[str],
        max_depth: int = 3,
        min_segments: int = 2,
    ):
        """Fit method for wise pizza to get wise pizza results and frequentist analysis
        of segments.

        Args:
            treatment (int, optional): name of variant at consideration. Defaults to 1.
            segments (List, optional): categorical segments to consider.
                Defaults to List[str].
            max_depth (int, optional): wise pizza parameter for max number of
                segment combinations.
                Defaults to 2.
            min_segments (int, optional): wise pizza parameter for minimal
            no. of segments to display. Defaults to 3.

        Returns:
            pd.DataFrame: frequentist result table for wise pizza selected segments
        """
        # prepare the dough (i.e. dataframe) of a wise pizza
        df_wp = self._wise_pizza_prep(treatment=treatment, segments=segments)

        # run Wise Pizza
        sf = explain_levels(
            df=df_wp,
            dims=segments,
            total_name=target,
            size_name=self.ed.variant,
            max_depth=max_depth,
            min_segments=min_segments,
        )
        self.sf = sf
        wp_fig = plot_segments(sf, width=300, height=300, return_fig=True)
        wp_segments = [s["segment"] for s in sf.segments]

        segment_results = dict()
        for seg in wp_segments:
            query_rows = pd.Series(True, index=range(len(self.ed.data)))
            for k, v in seg.items():
                query_rows = (query_rows) & (self.ed.data[k] == v)
            ed_segment = copy.deepcopy(self.ed)
            ed_segment.data = self.ed.data[query_rows]
            ed_segment.n_variants = 2
            segment_results[json.dumps(seg)] = compute_frequentist_results(ed_segment)

        return segment_results, wp_fig

    def _wise_pizza_prep(self, treatment: int = 1, segments: [] = List[str]):
        ps_counts = (
            self.ed.data.loc[self.ed.data[self.ed.variant].isin([0, treatment]), :]
            .groupby(by=segments)[self.ed.variant]
            .agg("count")
        )

        df_wp_prep = (
            self.ed.data.loc[self.ed.data[self.ed.variant].isin([0, treatment]), :]
            .groupby(by=segments + [self.ed.variant])[list(self.ed.targets)]
            .agg(["mean"])
        )

        # df_wp_prep[self.ed.variant] = df_wp_prep.join(ps_counts, on=segments)

        df_test = df_wp_prep.query(f"{self.ed.variant} == {treatment}").reset_index(
            level=[self.ed.variant]
        )
        df_ctrl = df_wp_prep.query(f"{self.ed.variant} == 0").reset_index(
            level=[self.ed.variant]
        )

        df_combined = self._te_per_segment_naive(df_test, df_ctrl)

        df_wp = df_combined.merge(
            ps_counts,
            left_index=True,
            right_index=True,
        )
        df_wp.reset_index(inplace=True)

        for m in self.ed.targets:
            df_wp[m] = df_wp[m] * df_wp[self.ed.variant]

        self.df_wp = df_wp

        return df_wp

    def get_sf(self):
        assert (
            self.sf is not None
        ), "You need to run wise_pizza_frequentist (fit) first!"
        return self.sf

    def _te_per_segment_naive(self, df_test, df_ctrl):
        # to be replaced by causaltune
        df_combined = df_test.merge(
            df_ctrl, left_index=True, right_index=True, suffixes=("_treat", "_ctrl")
        )

        for m in self.ed.targets:
            df_combined[m] = df_combined[f"{m}_treat"] - df_combined[f"{m}_ctrl"]
        df_combined = df_combined[list(self.ed.targets)]
        df_combined.columns = df_combined.columns.droplevel(level=1)

        return df_combined


class Segmentation:
    def __init__(self, ed, alpha: Optional[float] = 0.05) -> None:
        self.ed = ed
        self.df_wp = None
        self.sf = None
        self.alpha = alpha

    def frequentist_by_segment(self, category: str, max_segments: int = 7):
        assert category in self.ed.data.columns

        segments = (
            self.ed.data.groupby(category)
            .count()
            .sort_values(by=[self.ed.variant], ascending=False)
            .index.values.tolist()[:max_segments]
        )
        segment_results = {}
        for segment in segments:
            ed_segment = copy.deepcopy(self.ed)
            ed_segment.data = ed_segment.data[ed_segment.data[category] == segment]
            ft_segment = FrequentistTest(ed=ed_segment, alpha=self.alpha)
            ft_segment.compute()
            segment_results[segment] = ft_segment.get_results_table()

        results_table = pd.concat(
            segment_results.values(), axis=0, keys=segment_results.keys()
        )
        results_table.index.names = [category] + segment_results[
            segments[0]
        ].index.names[:]

        results_table = results_table.swaplevel(0, 2).sort_index()

        warnings.warn(
            "There is no p-value correction for segmentation analsysis implemented at"
            " this point."
        )

        return results_table

    def wise_pizza_frequentist(
        self,
        target: str,
        treatment: int = 1,
        segments=List[str],
        max_depth: int = 2,
        min_segments: int = 3,
        auto_display_df=True,
    ):
        """Fit method for wise pizza to get wise pizza results and frequentist analysis
        of segments.

        Args:
            treatment (int, optional): name of variant at consideration.
                Defaults to 1.
            segments (List, optional): categorical segments to consider.
                Defaults to List[str].
            max_depth (int, optional): wise pizza parameter for max
                number of segment combinations. Defaults to 2.
            min_segments (int, optional): wise pizza parameter for minimal
                no. of segments to display. Defaults to 3.

        Returns:
            pd.DataFrame: frequentist result table for wise pizza selected segments
        """
        # prepare the dough (i.e. dataframe) of a wise pizza
        df_wp = self._wise_pizza_prep(treatment=treatment, segments=segments)

        # run Wise Pizza
        sf = explain_levels(
            df=df_wp,
            dims=segments,
            total_name=target,
            size_name=self.ed.variant,
            max_depth=max_depth,
            min_segments=min_segments,
        )
        self.sf = sf

        wp_segments = [s["segment"] for s in sf.segments]

        segment_results = {}
        for seg in wp_segments:
            query_rows = pd.Series(True, index=range(len(self.ed.data)))
            for k, v in seg.items():
                query_rows = (query_rows) & (self.ed.data[k] == v)
            ed_segment = copy.deepcopy(self.ed)
            ed_segment.data = self.ed.data[query_rows]
            ed_segment.n_variants = 2
            ft_segment = FrequentistTest(ed=ed_segment, alpha=self.alpha)
            ft_segment.compute()
            segment_results[json.dumps(seg)] = ft_segment.get_results_table()
            results_table = pd.concat(
                segment_results.values(), axis=0, keys=segment_results.keys()
            )

        results_table.index.names = ["Pizza slices", "Variants", "Outcomes"]
        results_table = results_table.swaplevel(0, 2)
        results_table = results_table.sort_index()
        if auto_display_df:
            display(
                results_table.style.apply(highlight, axis=1)
                .set_table_attributes("style='display:inline'")
                .bar(subset=["Estimated_Effect_relative"], color="grey")
                .format(precision=3)
            )
        return results_table

    def _wise_pizza_prep(self, treatment: int = 1, segments=List[str]):
        ps_counts = (
            self.ed.data.loc[self.ed.data[self.ed.variant].isin([0, treatment]), :]
            .groupby(by=segments)[self.ed.variant]
            .agg("count")
        )

        df_wp_prep = (
            self.ed.data.loc[self.ed.data[self.ed.variant].isin([0, treatment]), :]
            .groupby(by=segments + [self.ed.variant])[list(self.ed.targets)]
            .agg(["mean"])
        )

        # df_wp_prep[self.ed.variant] = df_wp_prep.join(ps_counts, on=segments)

        df_test = df_wp_prep.query(f"{self.ed.variant} == {treatment}").reset_index(
            level=[self.ed.variant]
        )
        df_ctrl = df_wp_prep.query(f"{self.ed.variant} == 0").reset_index(
            level=[self.ed.variant]
        )

        df_combined = self._te_per_segment_naive(df_test, df_ctrl)

        df_wp = df_combined.merge(
            ps_counts,
            left_index=True,
            right_index=True,
        )
        df_wp.reset_index(inplace=True)

        for m in self.ed.targets:
            df_wp[m] = df_wp[m] * df_wp[self.ed.variant]

        self.df_wp = df_wp

        return df_wp

    def get_sf(self):
        assert (
            self.sf is not None
        ), "You need to run wise_pizza_frequentist (fit) first!"
        return self.sf

    def _te_per_segment_naive(self, df_test, df_ctrl):
        # to be replaced by autocausality
        df_combined = df_test.merge(
            df_ctrl, left_index=True, right_index=True, suffixes=("_treat", "_ctrl")
        )

        for m in self.ed.targets:
            df_combined[m] = df_combined[f"{m}_treat"] - df_combined[f"{m}_ctrl"]
        df_combined = df_combined[list(self.ed.targets)]
        df_combined.columns = df_combined.columns.droplevel(level=1)

        return df_combined



from tw_experimentation.utils import ExperimentDataset


from tw_experimentation.checker import (
    Monitoring,
    SegmentMonitoring,
    SequentialTest,
    NormalityChecks,
)

from tw_experimentation.statistical_tests import compute_frequentist_results, run_cuped
from tw_experimentation.segmentation_frequentist import run_segmentation_analysis


from tw_experimentation.bayes.bayes_test import BayesTest


from typing import List, Optional
import pandas as pd
import pickle as pkl


def generate_results(
    df: pd.DataFrame,
    variant: str,
    targets: List[str],
    date_created: Optional[str] = None,
    pre_experiment_cols: Optional[List[str]] = None,
    segments: Optional[List[str]] = None,
    remove_outliers=True,
):
    """Compute all results for an experiment.

    Args:
        df (pd.DataFrame): dataframe with experiment data
        variant (str): variant name
        targets (List[str]): target names
        date_created (Optional[str], optional): timestamp column name.
            Defaults to None.
        pre_experiment_cols (Optional[List[str]], optional): pre-experiment
            outcome metrics.Defaults to None.
        segments (Optional[List[str]], optional): segments for monitoring and
            segmentation analysis.
            Defaults to None.
        remove_outliers (bool, optional): Whether to remove outliers
            from continuous variables.
            Defaults to True.

    Returns:
        Dict: Output plots, tables etc.
    """
    output = dict()

    # Data model setup
    ed = ExperimentDataset(
        data=df,
        variant=variant,
        targets=targets,
        date=date_created,
        pre_experiment_cols=pre_experiment_cols,
        segments=segments,
    )
    ed.preprocess_dataset(remove_outliers=remove_outliers)
    emd = ed.experiment_meta_data()
    output["emd"] = emd

    # Monitoring and integrity checks
    monitor = Monitoring(ed)
    output["monitor_results"] = monitor.create_tables_and_plots()

    if isinstance(segments, list) and len(segments) > 0:
        segment_monitor = SegmentMonitoring(ed, segments)
        CHI_SQUAERED_ALPHA = 0.05
        output["segment_monitor_results"] = segment_monitor.create_tables_and_plots(
            chi_squared_alpha=CHI_SQUAERED_ALPHA
        )
    else:
        output["segment_monitor_results"] = None

    nc = NormalityChecks(ed)
    ALPHA_NORMALITY_CHECK = 0.05
    output["nco"] = nc.create_results(alpha=ALPHA_NORMALITY_CHECK)

    sequential_test = SequentialTest(ed)
    _ = sequential_test.sequential_test_results()
    output["fig_sequential_test"] = sequential_test.fig_sequential_test()

    # Frequentist results
    output["frequentist_test_results"] = compute_frequentist_results(ed)
    if ed.pre_experiment_cols is not None and len(ed.pre_experiment_cols) > 0:
        output["cuped"] = run_cuped(ed)
    if isinstance(segments, list) and len(segments) > 0:
        output["segmentation_analysis"] = run_segmentation_analysis(
            ed, segments=ed.segments
        )

    # Bayes results
    bt = BayesTest(ed=ed)
    output["bayes_result"] = bt.compute_posterior()

    return output


def save_output(path: str, name: str, output: dict):
    """Save output to pickle file.

    Args:
        path (str): path name
        name (str): file name
        output (dict): output dict to save
    """
    with open(path + name + ".pkl", "wb") as f:
        pkl.dump(output, f)


def load_output(path: str, name: str):
    with open(path + name + ".pkl", "rb") as file:
        # Load the object from the file
        output = pkl.load(file)

    return output



from tw_experimentation.variance_reduction.variance_reduction_method import (
    VarianceReductionMethod,
)
from typing import List
import statsmodels.api as sm
import pandas as pd


class MultivariateRegression(VarianceReductionMethod):
    def fit(
        self,
        data: pd.DataFrame,
        treatment_column: str,
        target_column: str,
        covariate_columns: List[str],
        **kwargs,
    ):
        """Apply Multivariate Regression to data.

        Args:
            data (pd.DataFrame): experiment data
            treatment_column (str): name of column containing treatment flags
            target_column (str): name of column containing target metric
            covariate_columns (List[str]): list of names of covariate columns

        Returns:
            MultivariateRegression: self
        """

        self.treatment_column = treatment_column

        covariates = data[covariate_columns]
        target = data[target_column]
        treatment = data[treatment_column]

        # merge together treatment and covariates
        X = covariates.assign(T=treatment)

        # rename column T
        X = X.rename(columns={"T": treatment_column})

        # fit multivariate regression
        self.regression_results = sm.OLS(target, sm.add_constant(X)).fit()

        # fit baseline
        self.fit_baseline(
            data=data, treatment_column=treatment_column, target_column=target_column
        )

        # statistics
        self.estimate = self.regression_results.params[treatment_column]
        self.p_value = self.regression_results.pvalues[treatment_column]
        self.conf_int_95 = (
            self.regression_results.conf_int(alpha=0.05, cols=None)[0][
                treatment_column
            ],
            self.regression_results.conf_int(alpha=0.05, cols=None)[1][
                treatment_column
            ],
        )
        self.variance_reduction_rate = self.calculate_variance_reduction()

        return self

    def calculate_variance_reduction(self):
        """Calculate the variance reduction rate."""

        frac = (
            self.regression_results.bse[self.treatment_column]
            / self.baseline_results.bse[1]
        )

        return 1 - frac**2

    def ci_width(self, alpha: float = 0.05):
        """Calculate the `1-alpha * 100%` confidence interval."""

        return (
            self.regression_results.conf_int(alpha=alpha, cols=None)[1][
                self.treatment_column
            ]
            - self.regression_results.conf_int(alpha=alpha, cols=None)[0][
                self.treatment_column
            ]
        )

    def calculate_ci_width_reduction(self, alpha=0.05):
        """Calculate the width reduction rate in 1-`alpha` * 100% confidence
        interval."""

        frac = self.ci_width(self.regression_results, alpha) / self.ci_width(
            self.baseline_results, alpha
        )

        return 1 - frac


# =================================================================================================
# =================================================================================================
# =================================================================================================


class MultivariateRegressionAdjusted(VarianceReductionMethod):
    def fit(
        self,
        data: pd.DataFrame,
        treatment_column: str,
        target_column: str,
        covariate_columns: List[str],
        **kwargs,
    ):
        """Apply Multivariate Regression to data.

        Args:
            data (pd.DataFrame): experiment data
            treatment_column (str): name of column containing treatment flags
            target_column (str): name of column containing target metric
            covariate_columns (List[str]): list of names of covariate columns

        Returns:
            MultivariateRegression: self
        """

        self.treatment_column = treatment_column

        covariates = data[covariate_columns]
        target = data[target_column]
        treatment = data[treatment_column]

        # prepare adjustment
        X = covariates.copy()
        # for col in covariate_columns:
        #     X[col+'_tilde'] = treatment * (X[col] - X[col].mean())

        X = pd.concat(
            [X]
            + [
                (treatment * (X[col] - X[col].mean())).rename(col + "_tilde")
                for col in covariate_columns
            ],
            axis=1,
        )

        # merge together treatment and covariates
        X = X.assign(T=treatment)

        # rename column T
        X = X.rename(columns={"T": treatment_column})

        # fit multivariate regression
        self.regression_results = sm.OLS(target, sm.add_constant(X)).fit()

        # fit baseline
        self.fit_baseline(
            data=data, treatment_column=treatment_column, target_column=target_column
        )

        # statistics
        self.estimate = self.regression_results.params[treatment_column]
        self.p_value = self.regression_results.pvalues[treatment_column]
        self.conf_int_95 = (
            self.regression_results.conf_int(alpha=0.05, cols=None)[0][
                treatment_column
            ],
            self.regression_results.conf_int(alpha=0.05, cols=None)[1][
                treatment_column
            ],
        )
        self.variance_reduction_rate = self.calculate_variance_reduction()

        return self

    def calculate_variance_reduction(self):
        """Calculate the variance reduction rate."""

        frac = (
            self.regression_results.bse[self.treatment_column]
            / self.baseline_results.bse[1]
        )

        return 1 - frac**2

    def ci_width(self, alpha: float = 0.05):
        """Calculate the `1-alpha * 100%` confidence interval."""

        return (
            self.regression_results.conf_int(alpha=alpha, cols=None)[1][
                self.treatment_column
            ]
            - self.regression_results.conf_int(alpha=alpha, cols=None)[0][
                self.treatment_column
            ]
        )

    def calculate_ci_width_reduction(self, alpha=0.05):
        """Calculate the width reduction rate in 1-`alpha` * 100% confidence
        interval."""

        frac = self.ci_width(self.regression_results, alpha) / self.ci_width(
            self.baseline_results, alpha
        )

        return 1 - frac



from tw_experimentation.variance_reduction.variance_reduction_method import (
    VarianceReductionMethod,
)
from tw_experimentation.variance_reduction.utils import (
    split_dataframe,
)
from typing import List

import statsmodels.api as sm
import statsmodels.formula.api as smf
import pandas as pd
import numpy as np
from flaml import AutoML
from xgboost import XGBRegressor
from sklearn.model_selection import StratifiedKFold


class MLRATE(VarianceReductionMethod):
    """Implements Machine Learning Regression-Adjusted Treatment Effect Estimator
    (MLRATE)."""

    def fit(
        self,
        data: pd.DataFrame,
        K_splits: int,
        treatment_column: str,
        target_column: str,
        covariate_columns: List[str],
        model=AutoML,
        model_init_config: dict = {},
        model_fit_config: dict = {},
        **kwargs,
    ):
        """

        Args:
            data (pd.DataFrame): experiment data
            K_splits (int): number of cross-fitting folds/splits
            treatment_column (str): name of column containing the treatment flag
            target_column (str): name of column containing the target metric
            covariate_columns (List[str]): list of names of covariate columns
            model (_type_, optional): regression model. Defaults to AutoML.
            model_init_config (dict, optional): configuration parameters
                passed at the initialization of the model. Defaults to {}.
            model_fit_config (dict, optional): configuration parameters passed at
                the fitting of the model. Defaults to {}.

        Returns:
            MLRATE: self
        """

        # reset indexing
        data = data.reset_index(drop=True)  # TODO: check if overwrites data

        assert set(data[treatment_column].unique()) == {0, 1}

        # split data into `K_splits`-folds for cross-fitting
        splits, index_to_split_map = split_dataframe(df=data, K=K_splits)
        N = len(data)

        target = data[target_column]
        treatment = data[treatment_column]
        covariates = data[covariate_columns]

        # initialization
        g_ks = [model(**model_fit_config) for _ in range(K_splits)]
        g_k_scores = [None for _ in range(K_splits)]

        # cross-fitting the regressors
        # print('fitting gs')
        for k, split_indices in enumerate(splits):
            all_indices = np.array([i for i in range(N)])
            indices_complement = np.setdiff1d(all_indices, split_indices)

            X_train = covariates.iloc[indices_complement]
            y_train = target.iloc[indices_complement]

            # g_k = regressor.fit(X_train, y_train, **model_fit_config)
            # g_ks.append(g_k)

            g_ks[k].fit(X_train, y_train, **model_fit_config)
            g_k_scores[k] = g_ks[k].score(X_train, y_train)

        g_pred = np.zeros(N)  # test set predictions

        # predicting the target
        # print('predicting')
        for split_indices, g_k in zip(splits, g_ks):
            g_k_pred = g_k.predict(covariates.iloc[split_indices])
            g_pred[split_indices] = g_k_pred

        self._diag = {
            "g_ks": g_ks,
            "g_pred": g_pred,
            "g_k_scores": g_k_scores,
            "splits": splits,
            "index_to_split_map": index_to_split_map,
            "treatment": treatment,
            "target": target,
            "covariates": covariates,
        }

        g_bar = g_pred.mean()  # mean of test set predictionss

        # assembling dataframe for OLS
        # print('assembling df')
        ml_rate_columns = {
            "target": target,
            treatment_column: treatment,
            "g_pred": g_pred,
            "g_pred_difference": np.multiply(treatment, g_pred - g_bar),
        }

        ml_rate_df = pd.DataFrame(ml_rate_columns)

        # OLS estimator
        # print('regressing')
        self.regression_results = sm.OLS(
            ml_rate_df["target"],
            sm.add_constant(
                ml_rate_df[[treatment_column, "g_pred", "g_pred_difference"]]
            ),
        ).fit()
        self.fit_baseline(
            data=data, treatment_column=treatment_column, target_column=target_column
        )

        # generate statistics
        # print('calculating statistics')
        self.estimate = self.regression_results.params[treatment_column]
        self.p_value = self.regression_results.pvalues[treatment_column]
        self.conf_int_95 = (
            self.regression_results.conf_int(alpha=0.05, cols=None)[0][
                treatment_column
            ],
            self.regression_results.conf_int(alpha=0.05, cols=None)[1][
                treatment_column
            ],
        )
        self.variance_reduction_rate = self.calculate_variance_reduction()
        self.robust_variance_estimate = self.robust_variance_estimator(
            Y=target.to_numpy(), T=treatment.to_numpy(), g_pred=g_pred
        )
        return self

    # TODO: Fix! This does not seem to be working properly
    @staticmethod
    def robust_variance_estimator(Y: np.array, T: np.array, g_pred: np.array):
        """Compute the robust variance estimator.

        Args:
            Y (np.array): target metric vector
            T (np.array): treatment vector
            g_pred (np.array): vector of predictions of the target metric

        Returns:
            float: variance estimate
        """

        N = len(Y)

        p_hat = T.sum() / N

        Y_0 = Y[T == 0]
        Y_1 = Y[T == 1]

        var_term_0 = np.var(Y_0) / (1 - p_hat)
        var_term_1 = np.var(Y_1) / (p_hat)
        var_term_g = np.var(g_pred) / ((1 - p_hat) * p_hat)

        Z = np.column_stack([np.ones(N), T, g_pred, T * g_pred])

        beta_hat = np.linalg.inv(Z.T @ Z / N) @ (Y.T @ Z / N)

        beta_2_term = beta_hat[2]
        beta_3_term = beta_hat[3]

        sigma_hat_sqrd = (
            var_term_0
            + var_term_1
            - var_term_g
            * (beta_2_term * p_hat + (beta_2_term + beta_3_term) * (1 - p_hat))
        )

        return sigma_hat_sqrd

    def calculate_variance_reduction(self):
        """Calculate variance reduction for method.

        Returns:
            float: variance reduction rate
        """

        frac = (
            self.regression_results.bse[1] / self.baseline_results.bse[1]
        )  # NOTE: regular, not robust errors

        return 1 - frac**2

    @staticmethod
    def ci_width(ols_results, alpha=0.05):  # TODO: typing
        """Calculate the `1-alpha * 100%` confidence interval."""

        return (
            ols_results.conf_int(alpha=alpha, cols=None)[1][1]
            - ols_results.conf_int(alpha=alpha, cols=None)[0][1]
        )

    def calculate_ci_width_reduction(self, alpha: float = 0.05) -> float:
        """Calculate confidence interval width reduction."""

        frac = self.ci_width(self.regression_results, alpha) / self.ci_width(
            self.baseline_results, alpha
        )

        return 1 - frac


# ==============================================================================
# ==============================================================================
# ==============================================================================

# https://github.com/muratunalphd/Blog-Posts/blob/main/variance-reduction-methods/MLRATE.ipynb


class AltMLRATE(VarianceReductionMethod):
    def fit(
        self,
        data: pd.DataFrame,
        K_splits: int,
        treatment_column: str,
        target_column: str,
        covariate_columns: List[str],
        **kwargs,
    ):
        dfml = data.reset_index(drop=True)  # TODO: check if overwrites data

        XGB_reg = XGBRegressor(
            learning_rate=0.1, max_depth=6, n_estimators=500, reg_lambda=1
        )

        # X_mat = dfml.columns.tolist()[0:p]
        X_mat = covariate_columns

        kfold = StratifiedKFold(n_splits=2, shuffle=True, random_state=1)

        ix = []
        Yhat = []
        for train_index, test_index in kfold.split(dfml, dfml[treatment_column]):
            df_train = dfml.iloc[train_index].reset_index()
            df_test = dfml.iloc[test_index].reset_index()

            X_train = df_train[X_mat].copy()
            y_train = df_train[target_column].copy()
            X_test = df_test[X_mat].copy()
            XGB_reg.fit(X_train, y_train)

            Y_hat = XGB_reg.predict(X_test)

            ix.extend(list(test_index))
            Yhat.extend(list(Y_hat))

        df_ml = (
            pd.DataFrame({"ix": ix, "Yhat": Yhat})
            .sort_values(by="ix")
            .reset_index(drop=True)
        )
        df_ml[[target_column, treatment_column]] = dfml[
            [target_column, treatment_column]
        ]
        df_ml["Ytilde"] = df_ml["Yhat"] - np.mean(df_ml["Yhat"])
        df_ml["Yres"] = df_ml[target_column] - df_ml["Yhat"]
        df_ml = df_ml.drop("ix", axis=1)

        # mlrate =  smf.ols('Y ~ T + Yhat + T:Ytilde',
        #              data = df_ml).fit(cov_type='HC1',use_t=True)

        mlrate = smf.ols(
            target_column
            + " ~ "
            + treatment_column
            + " + Yhat + "
            + treatment_column
            + ":Ytilde",
            data=df_ml,
        ).fit(cov_type="HC1", use_t=True)

        self.regression_results = mlrate

        # fit difference-in-means estimator
        self.fit_baseline(
            data=data, treatment_column=treatment_column, target_column=target_column
        )

        # generate statistics
        # print('calculating statistics')
        self.estimate = self.regression_results.params[treatment_column]
        self.p_value = self.regression_results.pvalues[treatment_column]
        self.conf_int_95 = (
            self.regression_results.conf_int(alpha=0.05, cols=None)[0][
                treatment_column
            ],
            self.regression_results.conf_int(alpha=0.05, cols=None)[1][
                treatment_column
            ],
        )
        self.variance_reduction_rate = self.calculate_variance_reduction()
        return self

    def calculate_variance_reduction(self):
        """Calculate variance reduction for method.

        Returns:
            float: variance reduction rate
        """

        frac = (
            self.regression_results.bse[1] / self.baseline_results.bse[1]
        )  # NOTE: regular, not robust errors

        return 1 - frac**2

    @staticmethod
    def ci_width(ols_results, alpha=0.05):  # TODO: typing
        """Calculate the `1-alpha * 100%` confidence interval."""

        return (
            ols_results.conf_int(alpha=alpha, cols=None)[1][1]
            - ols_results.conf_int(alpha=alpha, cols=None)[0][1]
        )

    def calculate_ci_width_reduction(self, alpha: float = 0.05) -> float:
        """Calculate confidence interval width reduction."""

        frac = self.ci_width(self.regression_results, alpha) / self.ci_width(
            self.baseline_results, alpha
        )

        return 1 - frac



from snowflake.sqlalchemy import URL
from sqlalchemy import create_engine
from prettytable import PrettyTable
import pandas as pd
import numpy as np
from typing import List, Generator


def create_pretty_table(
    rows: List[str], columns: List[str], contents: np.ndarray
) -> PrettyTable:  # TODO: verify typing
    """Return a pretty table from given row names, column names and contents."""

    # Create a PrettyTable object
    table = PrettyTable()
    table.field_names = [""] + columns

    # Populate the table with data
    for i, row in enumerate(rows):
        table.add_row([row] + list(contents[i]))

    return table


def query_snowflake(
    query: str,
    username: str,
    warehouse: str,
    database: str,
    account: str = "account_name",
    region: str = "region_name",
) -> pd.DataFrame:
    """Query a snowflake database."""

    authenticator = "EXTERNALBROWSER"

    engine = create_engine(
        URL(
            account=account,
            region=region,
            user=username,
            authenticator=authenticator,
            database=database,
            warehouse=warehouse,
        )
    )
    try:
        connection = engine.connect()
        df = pd.read_sql(query, connection)
    finally:
        connection.close()
        engine.dispose

    return df


def bootstrap_generator(
    data: pd.DataFrame, n_bootstrap: int, sample_size: float = False
) -> Generator[
    pd.DataFrame, None, None
]:  # TODO: typing of sample_size and verify return type
    """Generates bootstrap samples of data.

    Args:
        data (pd.DataFrame): data
        n_bootstrap (int): number of bootstrap samples
        sample_size (bool, optional): size of each bootstrap sample. Defaults to False.

    Yields:
        pd.DataFrame: Bootstrapped sample
    """

    if not sample_size:
        sample_size = len(data)

    for _ in range(n_bootstrap):
        yield data.sample(n=sample_size, replace=True)


# NOTE: Say the sample split is 0.3, i think that for some data the control will be
# 0.3 and for some it will be the treatment,
# depending on which will end up first when unique() is taken
def subsample_generator(
    data: pd.DataFrame,
    sample_size: int,
    n_bootstrap: int,
    treatment_col: str,
    sample_split=False,
) -> Generator[
    pd.DataFrame, None, None
]:  # TODO: typing of sample_split and verify generator
    """Generates subsamples of experiment data.

    Args:
        data (pd.DataFrame): experiment data
        sample_size (int): sample size of each subsample
        n_bootstrap (int): number of subsamples
        treatment_col (str): name of treatment column
        sample_split (bool, optional): custom sample split proportion.
            Defaults to False.

    Yields:
        pd.DataFrame: A subsample of experiment data.
    """

    unique_treatments = data[treatment_col].unique()

    assert len(unique_treatments) == 2

    # mask = data[treatment_col]==unique_treatments[0]
    # data_0 = data.loc[mask]
    # data_1 = data.loc[-mask]

    data_0 = data.loc[data[treatment_col] == unique_treatments[0]]
    data_1 = data.loc[data[treatment_col] == unique_treatments[1]]

    if not sample_split:
        sample_split = len(data_0) / len(data)
    n0 = int(sample_size * sample_split)
    n1 = sample_size - n0

    for bootstrapped_0, bootstrapped_1 in zip(
        bootstrap_generator(data_0, n_bootstrap, n0),
        bootstrap_generator(data_1, n_bootstrap, n1),
    ):
        yield pd.concat([bootstrapped_0, bootstrapped_1]).sample(frac=1).reset_index(
            drop=True
        )


def split_dataframe(df: pd.DataFrame, K: int) -> List[np.array]:
    """Splits a dataframe into K splits uniformly at random.

    Args:
        df (pd.DataFrame): data
        K (int): number of splits

    Returns:
        List[np.array]: list of indices of splits
    """

    # Shuffle the indices of the DataFrame
    shuffled_indices = np.random.permutation(df.index)

    # Calculate the approximate size of each split
    split_size = len(df) // K

    # Initialize an empty list to store the split indices
    split_indices_list = []
    index_to_split_map = {}

    # Split the DataFrame into K approximately equal-sized parts
    for i in range(K):
        start = i * split_size
        end = (i + 1) * split_size if i < K - 1 else len(df)

        # Extract the current split indices based on shuffled indices
        current_split_indices = shuffled_indices[start:end]

        # Append the split indices to the list
        split_indices_list.append(current_split_indices)

        for ind in current_split_indices:
            index_to_split_map[ind] = i

    return split_indices_list, index_to_split_map


def bootstrap_data(
    data: pd.DataFrame, n_bootstrap: int, sample_size: float = False
) -> np.array:
    """Returns the indices of bootstrap samples of data."""
    N = len(data)

    if not sample_size:
        sample_size = N

    # assert sample_size <= N

    # bootstrap_indices = np.zeros((n_bootstrap, sample_size), dtype=int)

    # for i in range(n_bootstrap):

    bootstrap_indices = np.random.choice(
        data.index, size=(n_bootstrap, sample_size), replace=True
    )

    return bootstrap_indices


def subsample_data(
    data: pd.DataFrame,
    n_bootstrap: int,
    treatment_column: str,
    sample_size: int = False,
    sample_split=False,
) -> np.array:
    """Returns the indices of subsamples of experiment data."""

    N = len(data)

    if not sample_size:
        sample_size = N

    assert set(data[treatment_column].unique()) == {0, 1}

    # mask = data[treatment_col]==unique_treatments[0]
    # data_0 = data.loc[mask]
    # data_1 = data.loc[-mask]

    data_0 = data.loc[data[treatment_column] == 0]
    data_1 = data.loc[data[treatment_column] == 1]

    if not sample_split:
        sample_split = len(data_0) / len(data)
    n0 = int(sample_size * sample_split)
    n1 = sample_size - n0

    bootstrap_indices_0 = bootstrap_data(data_0, n_bootstrap, n0)
    bootstrap_indices_1 = bootstrap_data(data_1, n_bootstrap, n1)

    return np.concatenate([bootstrap_indices_0, bootstrap_indices_1], axis=1)


def aaify(data: pd.DataFrame, treatment_column: str, frac_control=False):
    """Turn an A/B experiment data into A/A experiment data by randomly assigning
    treatment to control.

    Args:
        data (pd.DataFrame): experiment data
        treatment_column (str): name of column containing treatment flags
        frac_control (bool, optional): fraction of data to be assigned to control.
            Defaults to False. # TODO: typing

    Returns:
        pd.DataFrame: experiment data with treatment filtered out
            and new treatment assignment
    """

    aa_df = data.loc[data[treatment_column] == 0]

    N = len(aa_df)

    p = frac_control if frac_control else 0.5

    treatment_assignment = np.random.binomial(1, p, size=N)

    aa_df[treatment_column] = treatment_assignment

    return aa_df.reset_index(drop=True)


def add_synthetic_effect(
    data: pd.DataFrame,
    treatment_column: str,
    target_column: str,
    effect_size: float,
):
    """Add a synthetic effect to the experiment data.

    Args:
        data (pd.DataFrame): experiment data
        treatment_column (str): name of column containing treatment flags
        effect_size (float): effect size
        treatment_column (str): name of column containing effect

    Returns:
        pd.DataFrame: experiment data with synthetic effect
    """

    synthetic_effect_data = data.copy()

    synthetic_effect_data[target_column] = (
        synthetic_effect_data[target_column]
        + synthetic_effect_data[treatment_column] * effect_size
    )

    return synthetic_effect_data


def diff_in_means(data: pd.DataFrame, treatment_column: str, target_column: str):
    """Calculate the difference in means between treatment and control groups.

    Args:
        data (pd.DataFrame): experiment data
        treatment_column (str): name of column containing treatment flags
        target_column (str): name of column containing target variable

    Returns:
        float: difference in means
    """

    treatment_mean = data.loc[data[treatment_column] == 1, target_column].mean()
    control_mean = data.loc[data[treatment_column] == 0, target_column].mean()

    return treatment_mean - control_mean



from tw_experimentation.constants import MetricType

import pandas as pd
import numpy as np


from typing import List

import numpyro
import numpyro.distributions as dist


from jax import numpy as jnp


from tw_experimentation.bayes.bayes_params import (
    DEFAULT_LIKELIHOOD_MODELS,
    DEFAULT_PRIOR_MODELS,
    DEFAULT_PRIOR_MODEL_PARAMS,
)


class BayesModel:
    def __init__(
        self,
        metric_type: str,
        observations: pd.core.groupby.generic.DataFrameGroupBy,
        n_variants: int = 2,
        do_bayesian_updating: bool = True,
    ) -> None:
        self.metric_type = metric_type
        self.do_bayesian_updating = do_bayesian_updating
        self.observations = {}
        for variant_name, variant_observations in observations:
            self.observations[str(variant_name)] = variant_observations.to_numpy()

        self.n_variants = n_variants

        assert len(self.observations) == self.n_variants

        available_types = [metric_type.value for metric_type in MetricType]
        assert (
            metric_type in available_types
        ), "Metric types must be one of 'binary', 'discrete', or 'continuous'"

        self.likelihood_model = DEFAULT_LIKELIHOOD_MODELS[self.metric_type]

        self.prior_model = DEFAULT_PRIOR_MODELS[self.metric_type]
        self.prior_model_params = DEFAULT_PRIOR_MODEL_PARAMS[self.metric_type]

    def set_model(
        self,
        likelihood_model,
        variables: List[str],
        prior_models: List,
        params_models: List[dict],
    ):
        self.set_likelihood_model(likelihood_model)

        # reset prior_modle and prior_model_params
        self.prior_model = {}
        self.prior_model_params = {}

        for model_and_params in zip(variables, prior_models, params_models):
            self.set_prior_model(*model_and_params)

        assert self._model_is_well_defined()

    def set_prior_model(self, variable: str, model, model_params: dict):
        """Set prior model for a likelihood model parameter.

        Args:
            variable (str): name of variab model variable to be fed into likelihood
            model (...): numpyro distribution used as prior
            model_params (dict): args and kwargs for prior model
                                Must be of form
                                {'args': (a,b),
                                 'kwargs': {'c': d}
                                 }
        """
        self.prior_model[variable] = model
        self.prior_model_params[variable] = model_params

    def set_prior_model_param(self, variable: str, model_params: dict):
        """Set parameters for prior.

        Args:
            variable (str):     name of variab model variable to be fed into likelihood
            model_params (dict): args and kwargs for prior model
                                Must be of form
                                {'args': (a,b),
                                 'kwargs': {'c': d}
                                 }
        """
        self.prior_model_params[variable] = model_params

    def set_likelihood_model(self, model):
        self.likelihood_model = model

    def build_model(self):
        models = {}
        deltas = {}
        ate = {}
        means = {}
        aux_dist = {}

        for k, obs in self.observations.items():
            models[k] = {}

            if int(k) > 0:  # treatment to control comparison
                deltas[k] = {}

            for likelihood_param, distribution in self.prior_model.items():
                models[k][likelihood_param] = numpyro.sample(
                    f"variant_{k}_{likelihood_param}",
                    distribution(
                        *self.prior_model_params[likelihood_param]["args"],
                        **self.prior_model_params[likelihood_param]["kwargs"],
                    ),
                )

                if int(k) > 0:  # treatment to control comparison
                    deltas[k][likelihood_param] = numpyro.deterministic(
                        f"delta_{k}_{likelihood_param}",
                        models[k][likelihood_param] - models["0"][likelihood_param],
                    )

            if self.likelihood_model.auxiliary_zero_inflation:
                aux_dist[k] = numpyro.sample(
                    f"aux_dist_{k}", dist.Uniform(low=0, high=1)
                )
                if self.do_bayesian_updating:
                    with numpyro.plate(f"likelihood_{k}", obs[np.nonzero(obs)].size):
                        numpyro.sample(
                            f"likelihood_model_{k}",
                            self.likelihood_model(**models[k]),
                            obs=obs[np.nonzero(obs)],
                        )
                    with numpyro.plate(f"posterior_gate_{k}", obs.size):
                        numpyro.sample(
                            f"posterior_gate_model_{k}",
                            dist.Bernoulli(probs=aux_dist[k]),
                            obs=jnp.where(obs > 0, 1, 0),
                        )

            else:
                if self.do_bayesian_updating:
                    with numpyro.plate(f"likelihood_{k}", obs.size):
                        numpyro.sample(
                            f"likelihood_model_{k}",
                            self.likelihood_model(**models[k]),
                            obs=obs,
                        )

            if int(k) > 0:
                # ate[k] = numpyro.deterministic(
                #     f"ate_{k}",
                #     self.posterior_ate(
                #         models,
                #         k,
                #         auxiliary_zero_inflation=self.likelihood_model.auxiliary_zero_inflation,
                #         aux_dist=aux_dist,
                #     ),
                # )

                ate[k] = numpyro.deterministic(
                    f"ate_{k}",
                    self.posterior_ate(
                        models,
                        k,
                        auxiliary_zero_inflation=self.likelihood_model.auxiliary_zero_inflation,  # noqa: E501
                        aux_dist=aux_dist,
                    ),
                )

            means[k] = numpyro.deterministic(
                f"means_{k}",
                self.posterior_means(
                    models,
                    k,
                    auxiliary_zero_inflation=self.likelihood_model.auxiliary_zero_inflation,  # noqa: E501
                    aux_dist=aux_dist,
                ),
            )

    def posterior_ate(
        self, numpyro_model, treatment, auxiliary_zero_inflation=False, aux_dist=None
    ):
        if auxiliary_zero_inflation:
            assert aux_dist is not None

        add_zeros_t = aux_dist[treatment] if auxiliary_zero_inflation else 1
        add_zeros_c = aux_dist["0"] if auxiliary_zero_inflation else 1
        if self.metric_type == "binary":
            ate = numpyro_model[treatment]["probs"] - numpyro_model["0"]["probs"]
        elif self.metric_type == "continuous":
            ate = (
                jnp.exp(
                    numpyro_model[treatment]["loc"]
                    + numpyro_model[treatment]["scale"] ** 2 / 2
                )
                * add_zeros_t
            ) - (
                jnp.exp(
                    numpyro_model["0"]["loc"] + numpyro_model["0"]["scale"] ** 2 / 2
                )
                * add_zeros_c
            )
        elif self.metric_type == "discrete":
            ate = numpyro_model[treatment]["rate"] * numpyro_model[treatment][
                "gate"
            ] - (numpyro_model["0"]["rate"] * numpyro_model["0"]["gate"])
        return ate

    def posterior_means(
        self, numpyro_model, variant, auxiliary_zero_inflation=False, aux_dist=None
    ):
        add_zeros = aux_dist[variant] if auxiliary_zero_inflation else 1
        if self.metric_type == "binary":
            return numpyro_model[variant]["probs"]
        elif self.metric_type == "continuous":
            return (
                jnp.exp(
                    numpyro_model[variant]["loc"]
                    + numpyro_model[variant]["scale"] ** 2 / 2
                )
                * add_zeros
            )
        elif self.metric_type == "discrete":
            return numpyro_model[variant]["rate"] * numpyro_model[variant]["gate"]

    def _model_is_well_defined(self):
        # check whether parameters needed for likelihood model all
        # appear in self.prior_model
        return True

    def get_likelihood_variables(self):
        return list(self.prior_model.keys())



import pandas as pd
import numpy as np
import plotly.express as px
from plotly import graph_objects as go
from plotly.subplots import make_subplots
from functools import wraps

from tw_experimentation.utils import (
    variant_name_map,
    ExperimentDataset,
    variant_color_map,
    variantname_color_map,
    hex_to_rgb,
)
from typing import List, Dict, Union, Optional

import statsmodels.api as sm
import matplotlib.pyplot as plt

from scipy.stats.distributions import chi2

MAX_N_POINTS = 8000


# TODO: Potentially replace by https://github.com/predict-idlab/plotly-resampler/
def plotly_reduce_n_points_per_trace(
    fig: go.Figure, max_n_points: int = 5000, min_n_points_per_trace: int = 200
):
    """Reduces the number of data points in a Plotly figure to improve memory usage.

    Args:
        fig (plotly.graph_objs.Figure): The Plotly figure to be modified.
        max_n_points (int, optional): The maximum number of data points to keep
            in the figure. Defaults to 5000.

    Returns:
        plotly.graph_objs.Figure: The modified Plotly figure with reduced data points.
    """
    traces = fig.data
    n_traces = len(traces)
    max_points_per_trace = int(max_n_points / n_traces)
    for trace in traces:
        trace_length = len(trace.x)
        step = max(1, trace_length // max(min_n_points_per_trace, max_points_per_trace))
        selected_indices = list(
            range(0, trace_length, step)
        )  # Include the first element

        # Ensure the last element is included
        if selected_indices[-1] != trace_length - 1:
            selected_indices.append(trace_length - 1)
        if (
            hasattr(trace, "x")
            and hasattr(trace, "y")
            and trace.x is not None
            and trace.y is not None
            and trace.mode != "lines"
        ):
            trace.x = trace.x[selected_indices]
            trace.y = trace.y[selected_indices]

    return fig


def plotly_light_memory(max_n_points=5000, min_n_points_per_trace=200):
    """Decorator that reduces the number of points in a Plotly
        figure to optimize memory usage.

    Args:
        max_n_points (int, optional): The maximum number of points allowed
            in the figure. Defaults to 5000.
        min_n_points_per_trace (int, optional): The minimum number of points
            allowed per trace in the figure. Defaults to 200.

    Returns:
        function: The decorated function that returns a Plotly figure with reduced
            number of points per trace.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            fig = func(*args, **kwargs)
            return plotly_reduce_n_points_per_trace(
                fig,
                max_n_points=max_n_points,
                min_n_points_per_trace=min_n_points_per_trace,
            )

        return wrapper

    return decorator


def fig_variant_segment_dependence(
    chi_squared_table: pd.DataFrame,
    ed: ExperimentDataset,
    text_auto=False,
):
    """Heatmap for chi-squared test of dependence between variant and segment Normalises
    the color by setting the midpoint of the colorscale to the 95% quantile of the chi-
    squared distribution divided by degrees of freedom. Hence, a cell achieves above a
    heat at the colorscale midpoint if all cells had the same value, then the test has a
    p-value of .05.

    Args:
        chi_squared_table (pd.DataFrame): table of N(0,1) distributed statistics
        ed (ExperimentDataset): ExperimentDataset object containing experiment data
        color_index (int, optional): index of color scale to use. Defaults to 1.
        text_auto (bool, optional): whether to display cell values. Defaults to False.

    Returns:
        plotly.graph_objs._figure.Figure: plotly figure
    """
    INVERSE_CDF_THRESHOLD = 0.95
    deg_of_fr = (chi_squared_table.shape[0] - 1) * (chi_squared_table.shape[1] - 1)
    norm_constant = chi2.ppf(INVERSE_CDF_THRESHOLD, deg_of_fr)
    fig = px.imshow(
        chi_squared_table,
        title=(
            " Degree of observed dependence of segment"
            f" {chi_squared_table.axes[1].name} and variant pairs"
        ),
        text_auto=text_auto,
        color_continuous_scale="Reds",
        color_continuous_midpoint=norm_constant / deg_of_fr,
    )

    fig = fig.update_layout(
        yaxis=dict(
            tickmode="array",
            tickvals=list(range(ed.n_variants)),
            ticktext=list(variant_name_map(ed.n_variants).values()),
        )
    )
    # fig.update_layout(
    #     coloraxis={"colorscale": c.COLORSCALES[color_index % len(c.COLORSCALES)]}
    # )
    return fig


def plot_sample_size_pie(ed: ExperimentDataset):
    """Plots a pie chart showing the sample size per variant in an ExperimentDataset.

    Args:
        ed (ExperimentDataset): An ExperimentDataset object containing the data
            to be plotted.

    Returns:
        fig: A plotly figure object representing the pie chart.
    """
    data = pd.DataFrame(
        {
            "Variant": variant_name_map(ed.n_variants).values(),
            "Value": list(ed.sample_sizes.values()),
        }
    )
    color_map = variantname_color_map(list(variant_name_map(ed.n_variants).values()))
    fig = px.pie(
        data,
        names="Variant",
        values="Value",
        title="Sample size per variant",
        color="Variant",
        color_discrete_map=color_map,
    )
    fig.update_traces(textinfo="percent+value")
    return fig


@plotly_light_memory(max_n_points=MAX_N_POINTS)
def plot_dynamic_sample_size(df_dynamic_sample: pd.DataFrame, ed: ExperimentDataset):
    """Plots the sample size per variant over time.

    Args:
        df_dynamic_sample (pd.DataFrame): A DataFrame containing the sample size data.
        ed (ExperimentDataset): An ExperimentDataset object containing information
            about the experiment.

    Returns:
        A plotly.graph_objs._figure.Figure Plot with sample size evolution.
    """
    fig = px.line(
        df_dynamic_sample,
        x=ed.date,
        y="variant_cnt",
        color=ed.variant,
        color_discrete_sequence=variant_color_map(ed.n_variants),
        title="Sample size per variant over time",
    )
    variant_names = variant_name_map(ed.n_variants)
    fig.for_each_trace(lambda trace: trace.update(name=variant_names[int(trace.name)]))
    return fig


@plotly_light_memory(max_n_points=MAX_N_POINTS)
def target_metric_distribution(
    ed: ExperimentDataset, target: str, use_log: bool = False
):
    fig = go.Figure()
    for i in range(ed.n_variants):
        fig.add_trace(
            go.Histogram(
                x=ed.data.loc[
                    ed.data[ed.variant] == i,
                    target,
                ],
                name=variant_name_map(ed.n_variants)[i],
                marker=dict(color=variant_color_map(ed.n_variants)[i]),
            ),
        )
    fig.update_layout(
        title_text="Distribution of " + target + " per variant",
        xaxis_title_text=target,
        yaxis_title_text="Count",
    )

    if ed.metric_types[target] in ["continuous", "discrete"]:
        for perc in [80, 95, 99]:
            percentile = np.percentile(
                ed.data[ed.data[ed.variant] == 0][target],
                perc,
            )
            fig.add_vline(
                x=percentile,
                line_dash="dash",
                line_color="grey",
                annotation=dict(text=f"{perc}th percentile"),
                annotation_position="top right",
            )
        if use_log:
            fig.update_layout(yaxis_type="log")
    return fig


@plotly_light_memory(max_n_points=MAX_N_POINTS)
def plot_target_metric_cdf(ed: ExperimentDataset, target: str, use_log: bool = False):
    """Plots the cumulative density function (CDF) for a given target variable."""
    fig = px.ecdf(
        ed.data,
        x=target,
        color=ed.variant,
        title="Cumulative density function (CDF)",
        color_discrete_map=variant_color_map(ed.n_variants),
        log_x=(
            use_log if ed.metric_types[target] in ["continuous", "discrete"] else False
        ),
    )
    fig.for_each_trace(
        lambda trace: trace.update(
            name=variant_name_map(ed.n_variants)[int(trace.name)]
        )
    )
    return fig


@plotly_light_memory(max_n_points=MAX_N_POINTS)
def plot_qq_variants(
    qq_variants: Dict[str, np.ndarray], ed: ExperimentDataset, target: str
) -> go.Figure:
    """Plots a Q-Q plot for a given target between different variants.

    Args:
        qq_variants (Dict[str, np.ndarray]): A dictionary containing
            the quantiles  for each variant.
        ed (ExperimentDataset): An ExperimentDataset object
            containing the data for the experiment.
        target (str): The target variable to plot.

    Returns:
        go.Figure: A Plotly figure object containing the Q-Q plot.
    """
    # TODO: Add identity line
    color_per_variant = variant_color_map(ed.n_variants)
    fig = go.Figure()
    for variant in range(1, ed.n_variants):
        fig.add_trace(
            go.Scatter(
                x=qq_variants[variant], y=qq_variants[0], mode="markers", name=variant
            )
        )
    fig.for_each_trace(
        lambda trace: trace.update(
            name=variant_name_map(ed.n_variants)[int(trace.name)],
            marker=dict(color=color_per_variant[int(trace.name)]),
        )
    )
    fig.add_trace(
        go.Scatter(
            x=ed.data[target],
            y=ed.data[target],
            mode="lines",
            line=dict(color="#636efa"),
            name="Identity Line",
        )
    )

    fig.update_layout(
        title=f"Quantile-quantile plot for {target} between variants",
        xaxis_title="Control",
        yaxis_title="Treatment",
    )
    return fig


@plotly_light_memory(max_n_points=MAX_N_POINTS)
def plot_segment_sample_size(
    ed: ExperimentDataset,
    df_dyn_segments: pd.DataFrame,
    segment: List[str],
):
    color_per_variant = variant_color_map(ed.n_variants)
    fig = px.line(
        df_dyn_segments,
        x=ed.date,
        y="variant_cnt",
        color=ed.variant,
        facet_row=segment,
        title=f"sample size per category in segment {segment}",
        color_discrete_map=color_per_variant,
    )
    fig.for_each_trace(
        lambda trace: trace.update(
            name=variant_name_map(ed.n_variants)[int(trace.name)]
        )
    )
    return fig


def plot_segment_histograms(
    ed: ExperimentDataset,
    segment: List[str],
    biggest_segments: Optional[Union[None, List]] = None,
):
    color_per_variant = variant_color_map(ed.n_variants)
    variant_names = variant_name_map(ed.n_variants)

    if biggest_segments is not None:
        data = ed.data[ed.data[segment].isin(biggest_segments)]
    else:
        data = ed.data
    fig = go.Figure()

    for i in range(ed.n_variants):
        fig.add_trace(
            go.Histogram(
                x=data.loc[
                    data[ed.variant] == i,
                    segment,
                ],
                name=variant_names[i],
                marker=dict(color=color_per_variant[i]),
            ),
        )
    fig.update_layout(
        title_text="Distribution of " + segment + " per variant",
        xaxis_title_text=segment,
        yaxis_title_text="Count",
    )
    return fig


@plotly_light_memory(max_n_points=MAX_N_POINTS)
def plot_sequential_test(ed: ExperimentDataset, df_dyn_avg: pd.DataFrame):
    n_rows = len(ed.targets)

    variant_names = variant_name_map(ed.n_variants)
    color_per_variant = variant_color_map(ed.n_variants)

    fig = make_subplots(
        rows=n_rows,
        cols=3,
        row_titles=ed.targets,
        column_titles=["Metric average", "Treatment Effect", "p Value"],
    )

    for j in range(n_rows):
        fig_aux1 = px.line(
            df_dyn_avg,
            x=ed.date,
            y=f"{ed.targets[j]}_avg",
            color=ed.variant,
            color_discrete_map=color_per_variant,
        )
        fig_aux1.for_each_trace(
            lambda trace: trace.update(name=variant_names[int(trace.name)])
        )
        if j > 0:
            fig_aux1.update_traces(showlegend=False)
        for k in range(ed.n_variants):
            fig.add_trace(fig_aux1.data[k], row=j + 1, col=1)

        fig_aux2 = px.line(
            df_dyn_avg[df_dyn_avg[ed.variant] != 0],
            x=ed.date,
            y=f"{ed.targets[j]}_diff",
            color=ed.variant,
            color_discrete_map=color_per_variant,
        )
        fig_aux2.update_traces(showlegend=False)
        for k in range(1, ed.n_variants):
            fig.add_trace(fig_aux2.data[k - 1], row=j + 1, col=2)
            df_variant = df_dyn_avg[df_dyn_avg[ed.variant] == k]
            for bound in ["lower", "upper"]:
                fig.add_trace(
                    go.Scatter(
                        x=df_variant[ed.date],
                        y=df_variant[f"{ed.targets[j]}_CI_{bound}"],
                        fill="tonexty" if bound == "upper" else None,
                        showlegend=False,
                        line=dict(color=color_per_variant[k], width=0.3),
                        fillcolor=(
                            f"rgba{(*hex_to_rgb(color_per_variant[k][1:]), .2)}"
                            if bound == "upper"
                            else None
                        ),
                    ),
                    row=j + 1,
                    col=2,
                )

        fig_aux3 = px.line(
            df_dyn_avg[df_dyn_avg[ed.variant] != 0],
            x=ed.date,
            y=f"{ed.targets[j]}_p_values",
            color=ed.variant,
            color_discrete_map=color_per_variant,
        )
        fig_aux3.update_traces(showlegend=False)
        # fig_aux3.update_yaxes(range=[0, 1])
        for k in range(ed.n_variants - 1):
            fig.add_trace(fig_aux3.data[k], row=j + 1, col=3)

        fig.update_layout(
            title_text="Outcome metric monitor and sequential testing",
        )
        # fig.update_layout({"xaxis3": {"range": [0, 1.01]}})
    return fig


@plotly_light_memory(max_n_points=MAX_N_POINTS)
def plot_qq_normal(
    ed: ExperimentDataset,
    target: str,
    standardized_residuals: Dict[str, Dict[int, np.ndarray]],
) -> go.Figure:
    """Plots a quantile-quantile plot for a given target variable and its
        standardized residuals for each variant in the experiment.

    Args:
        ed (ExperimentDataset): The experiment dataset containing the data for
        each variant.
        target (str): The name of the target variable to plot.
        standardized_residuals (Dict[str, Dict[int, np.ndarray]]): A dictionary
            containing the standardized residuals for each variant
            and target variable.

    Returns:
        go.Figure: A Plotly figure object containing the quantile-quantile plot.
    """
    plt.ioff()
    fig = go.Figure()
    xdata = []
    ydata = []
    color_per_variant = variant_color_map(ed.n_variants)
    variant_names = variant_name_map(ed.n_variants)
    for variant in range(1, ed.n_variants):
        fig_mpl = sm.qqplot(standardized_residuals[target][variant], line="s")
        qqplot_data = fig_mpl.gca().lines
        xdata.append(qqplot_data[0].get_xdata())
        ydata.append(qqplot_data[0].get_ydata())

        fig.add_trace(
            {
                "type": "scatter",
                "x": qqplot_data[1].get_xdata(),
                "y": qqplot_data[0].get_ydata(),
                "mode": "markers",
                "name": variant_names[variant],
                "marker": dict(color=color_per_variant[variant]),
            }
        )
    all_data = xdata + ydata
    data_concat = np.concatenate(all_data)
    fig.add_trace(
        {
            "name": "identity line",
            "type": "scatter",
            "x": [data_concat.min(), data_concat.max()],
            "y": [data_concat.min(), data_concat.max()],
            "mode": "lines",
            "line": {"color": "#636efa"},
        }
    )

    fig["layout"].update(
        {
            "title": f"Quantile-Quantile Plot: {target}",
            "xaxis": {"title": "Theoretical Quantiles", "zeroline": False},
            "yaxis": {"title": "Sample Quantiles"},
        }
    )
    return fig



from tw_experimentation.utils import ExperimentDataset

from abc import ABC, abstractmethod

from tw_experimentation.variance_reduction.cuped import (
    CUPED,
    MultivariateCUPED,
)

from dataclasses import dataclass
from typing import Optional, Dict, Union, List

import numpy as np
import pandas as pd
from statsmodels.stats.weightstats import ttest_ind
from statsmodels.stats.proportion import (
    confint_proportions_2indep,
)
from statsmodels.stats.proportion import (
    test_proportions_2indep as ztest_proportions_2indep,
)
from scipy import stats as sps

import copy


# TODO: Write factory so that the frequentist test result can be
#   calculated for each segment?
@dataclass
class FrequentistPerTargetResults:
    variant_means: Dict[int, float]
    variant_stds: Dict[int, float]
    variant_n_samples: Dict[int, int]

    def __post_init__(self):
        self.n_treatments = len(self.variant_means.keys()) - 1

    def t_test_per_variant(self, direction="two-sided", pooled_variance: bool = True):
        p_values = dict()
        t_stats = dict()
        for variant in self.variant_means.keys():
            if variant > 0:
                t_stat, p_value = self.t_test(
                    variant, direction=direction, pooled_variance=pooled_variance
                )
                p_values[variant] = p_value
                t_stats[variant] = t_stat

        return p_values, t_stats

    def t_test_conf_interval_per_variant(
        self,
        direction="two-sided",
        pooled_variance: bool = True,
        alphas: Union[float, List[float]] = 0.05,
    ):
        if isinstance(alphas, float):
            alphas = [alphas] * len(self.variant_means.keys())
        else:
            assert isinstance(alphas, list) and len(alphas) == len(
                self.variant_means.keys()
            ), "alphas must be a list of length equal to the number of variants"

        cis_lower = dict()
        cis_upper = dict()
        for index, variant in enumerate(self.variant_means.keys()):
            if variant > 0:
                ci_lower, ci_upper = self.t_test_confidence_interval(
                    variant,
                    alpha=alphas[index],
                    direction=direction,
                    pooled_variance=pooled_variance,
                )
                cis_lower[variant] = ci_lower
                cis_upper[variant] = ci_upper
        return cis_lower, cis_upper

    def t_test(
        self, treatment, direction: str = "two-sided", pooled_variance: bool = True
    ):
        mean_treat = self.variant_means[treatment]
        mean_control = self.variant_means[0]
        std_treat = self.variant_stds[treatment]
        std_control = self.variant_stds[0]
        n_treat = self.variant_n_samples[treatment]
        n_control = self.variant_n_samples[0]

        se = self.t_test_standard_error(
            std_treat, std_control, n_treat, n_control, pooled_variance
        )

        t_stat = (mean_treat - mean_control) / se

        if direction == "two-sided":
            p_value = 2 * sps.t.cdf(-np.abs(t_stat), n_treat + n_control - 2)
        elif direction == "greater":
            p_value = sps.t.cdf(-np.abs(t_stat), n_treat + n_control - 2)
        elif direction == "less":
            p_value = sps.t.cdf(t_stat, n_treat + n_control - 2)
        return t_stat, p_value

    @staticmethod
    def t_test_standard_error(
        std_treat, std_control, n_treat, n_control, pooled_variance: bool = True
    ):
        if pooled_variance:
            var_pooled = (
                (n_treat - 1) * std_treat**2 + (n_control - 1) * std_control**2
            ) / (n_treat + n_control - 2)
            se = np.sqrt(var_pooled * (1 / n_treat + 1 / n_control))

        else:
            se = np.sqrt(std_treat**2 / n_treat + std_control**2 / n_control)
        return se

    def t_test_confidence_interval(
        self,
        treatment,
        direction: str = "two-sided",
        alpha: float = 0.05,
        pooled_variance: bool = True,
    ):
        mean_treat = self.variant_means[treatment]
        mean_control = self.variant_means[0]
        std_treat = self.variant_stds[treatment]
        std_control = self.variant_stds[0]
        n_treat = self.variant_n_samples[treatment]
        n_control = self.variant_n_samples[0]
        se = self.t_test_standard_error(
            std_treat, std_control, n_treat, n_control, pooled_variance
        )
        if direction == "two-sided":
            critical_value = sps.t.ppf(1 - alpha / 2, n_treat + n_control - 2)
            ci_lower = (mean_treat - mean_control) - critical_value * se
            ci_upper = (mean_treat - mean_control) + critical_value * se
        elif direction == "greater":
            ci_upper = np.inf
            critical_value = sps.t.ppf(1 - alpha, n_treat + n_control - 2)
            ci_lower = -critical_value * se + (mean_treat - mean_control)
        elif direction == "less":
            ci_lower = -np.inf
            critical_value = sps.t.ppf(alpha, n_treat + n_control - 2)
            ci_upper = critical_value * se + (mean_treat - mean_control)

        return ci_lower, ci_upper

    def relative_naive_treatment_effect(self, treatment):
        try:
            relative_te = (
                (self.variant_means[treatment] - self.variant_means[0])
                / self.variant_means[0]
                * 100
            )
        except ZeroDivisionError:
            relative_te = 0
        return relative_te

    def relative_naive_treatment_effect_per_variant(self):
        relative_tes = dict()
        for variant in self.variant_means.keys():
            if variant > 0:
                relative_tes[variant] = self.relative_naive_treatment_effect(variant)
        return relative_tes

    def absolute_naive_treatment_effect(self, treatment):
        return self.variant_means[treatment] - self.variant_means[0]

    def absolute_naive_treatment_effect_per_variant(self):
        absolute_tes = dict()
        for variant in self.variant_means.keys():
            if variant > 0:
                absolute_tes[variant] = self.absolute_naive_treatment_effect(variant)
        return absolute_tes

    def two_sample_proportion_test(
        self, treatment, direction="two-sided", pooled_variance: bool = True
    ):
        mean_treat = self.variant_means[treatment]
        mean_control = self.variant_means[0]
        n_treat = self.variant_n_samples[treatment]
        n_control = self.variant_n_samples[0]

        se = self.two_sample_proportion_standard_error(
            mean_treat, mean_control, n_treat, n_control, pooled_variance
        )
        z_stat = (mean_treat - mean_control) / se
        if direction == "two-sided":
            p_value = 2 * sps.norm.cdf(-np.abs(z_stat))
        elif direction == "greater":
            p_value = sps.norm.cdf(-np.abs(z_stat))
        elif direction == "less":
            p_value = sps.norm.cdf(z_stat)

        return z_stat, p_value

    @staticmethod
    def two_sample_proportion_standard_error(
        mean_treat, mean_control, n_treat, n_control, pooled_variance: bool = True
    ):
        sample_proportion = (mean_treat * n_treat + mean_control * n_control) / (
            n_treat + n_control
        )
        if pooled_variance:
            se = np.sqrt(
                sample_proportion
                * (1 - sample_proportion)
                * (1 / n_treat + 1 / n_control)
            )
        else:
            se = np.sqrt(
                mean_treat * (1 - mean_treat) / n_treat
                + mean_control * (1 - mean_control) / n_control
            )
        return se

    def two_sample_proportion_confidence_interval(
        self,
        treatment,
        alpha: float = 0.05,
        direction: str = "two-sided",
        pooled_variance: bool = True,
    ):
        mean_treat = self.variant_means[treatment]
        mean_control = self.variant_means[0]
        n_treat = self.variant_n_samples[treatment]
        n_control = self.variant_n_samples[0]
        se = self.two_sample_proportion_standard_error(
            mean_treat, mean_control, n_treat, n_control, pooled_variance
        )
        if direction == "two-sided":
            critical_value = sps.norm.ppf(1 - alpha / 2)
            ci_lower = (mean_treat - mean_control) - critical_value * se
            ci_upper = (mean_treat - mean_control) + critical_value * se
        elif direction == "greater":
            ci_upper = np.inf
            critical_value = sps.norm.ppf(1 - alpha)
            ci_lower = -critical_value * se + (mean_treat - mean_control)
        elif direction == "less":
            ci_lower = -np.inf
            critical_value = sps.norm.ppf(alpha)
            ci_upper = critical_value * se + (mean_treat - mean_control)

        return ci_lower, ci_upper

    def two_sample_proportion_confidence_interval_per_variant(
        self,
        direction="two-sided",
        pooled_variance: bool = True,
        alphas: Union[float, List[float]] = 0.05,
    ):
        if isinstance(alphas, float):
            alphas = [alphas] * len(self.variant_means.keys())
        else:
            assert isinstance(alphas, list) and len(alphas) == len(
                self.variant_means.keys()
            ), "alphas must be a list of length equal to the number of variants"
        cis_lower = dict()
        cis_upper = dict()
        for index, variant in enumerate(self.variant_means.keys()):
            if variant > 0:
                ci_lower, ci_upper = self.two_sample_proportion_confidence_interval(
                    variant,
                    alpha=alphas[index],
                    direction=direction,
                    pooled_variance=pooled_variance,
                )
                cis_lower[variant] = ci_lower
                cis_upper[variant] = ci_upper
        return cis_lower, cis_upper

    def two_sample_proportion_test_per_variant(
        self, direction="two-sided", pooled_variance: bool = True
    ):
        p_values = dict()
        z_stats = dict()
        for variant in self.variant_means.keys():
            if variant > 0:
                z_stat, p_value = self.two_sample_proportion_test(
                    variant, direction=direction, pooled_variance=pooled_variance
                )
                p_values[variant] = p_value
                z_stats[variant] = z_stat

        return p_values, z_stats


@dataclass
class FrequentistTestResults:
    target_variant_means: Dict[str, Dict[str, float]]
    target_variant_stds: Dict[str, Dict[str, float]]
    variant_n_samples: Dict[str, Dict[str, int]]
    target_metric_types: Dict[str, str]
    n_variants: int

    def __post_init__(self):
        self.n_targets = len(self.target_variant_means.keys())
        self.n_treatments = self.n_variants - 1
        self.stats_per_target = None
        self.CONTROL_LABEL = 0

    def compute_stats_per_target(
        self,
        direction: str = "two-sided",
        pooled_variance: bool = False,
        type_i_error: float = 0.05,
        multitest_correction: Optional[str] = None,
    ):
        """Compute statistical tests and confidence intervals for
            each target variable.

        Args:
            direction (str, optional): The direction of the test.
                Defaults to "two-sided".
                Options are ['two-sided', 'greater', 'less'].
            pooled_variance (bool, optional): Whether to use pooled
                variance for hypothesis test. Defaults to True.
            type_i_error (float, optional): The significance level(s)
                for confidence intervals. Defaults to 0.05.
            multitest_correction (str, optional): The method for multiple
                hypothesis correction. Defaults to None.
                Can be one of ['bonferroni', None].

        Returns:
            dict: A dictionary containing the computed statistical tests, p-values,
                confidence intervals, and treatment effects for each target variable.
        """
        # TODO: initiate multiple hypothesis correction here
        if multitest_correction is None:
            mhc = NoCorrection(type_i_error=type_i_error)
        elif multitest_correction == "bonferroni":
            mhc = BonferroniCorrection(
                type_i_error=type_i_error, n_treatments=self.n_treatments
            )
        else:
            raise NotImplementedError
        stats_per_target = dict()
        for target in self.target_variant_means.keys():
            fptr = FrequentistPerTargetResults(
                variant_means=self.target_variant_means[target],
                variant_stds=self.target_variant_stds[target],
                variant_n_samples=self.variant_n_samples,
            )
            relative_tes = fptr.relative_naive_treatment_effect_per_variant()
            absolute_tes = fptr.absolute_naive_treatment_effect_per_variant()
            if self.target_metric_types[target] in ["continuous", "discrete"]:
                p_values, t_stats = fptr.t_test_per_variant(
                    direction=direction, pooled_variance=pooled_variance
                )
                stats_per_target[target] = {
                    "test_stats": t_stats,
                    "p_values": p_values,
                    "test_type": "t_test",
                }
            elif self.target_metric_types[target] == "binary":
                p_values, z_stats = fptr.two_sample_proportion_test_per_variant(
                    direction=direction, pooled_variance=pooled_variance
                )
                stats_per_target[target] = {
                    "test_stats": z_stats,
                    "p_values": p_values,
                    "test_type": "proportion_test",
                }
            alphas = mhc.correct_alphas(p_values=p_values)
            assert isinstance(
                alphas, float
            ), "alphas must be a float, alternative not implemented yet"
            stats_per_target[target]["p_values"] = mhc.correct_p_values(
                p_values=p_values
            )
            stats_per_target[target]["absolute_tes"] = absolute_tes
            stats_per_target[target]["relative_tes"] = relative_tes

            if self.target_metric_types[target] in ["continuous", "discrete"]:
                cis_lower, cis_upper = fptr.t_test_conf_interval_per_variant(
                    direction=direction,
                    pooled_variance=pooled_variance,
                    alphas=alphas,
                )
                stats_per_target[target]["cis_lower"] = cis_lower
                stats_per_target[target]["cis_upper"] = cis_upper
            elif self.target_metric_types[target] == "binary":
                (
                    cis_lower,
                    cis_upper,
                ) = fptr.two_sample_proportion_confidence_interval_per_variant(
                    direction=direction,
                    pooled_variance=pooled_variance,
                    alphas=alphas,
                )
                stats_per_target[target]["cis_lower"] = cis_lower
                stats_per_target[target]["cis_upper"] = cis_upper
            else:
                raise NotImplementedError

            stats_per_target[target]["is_significant"] = {
                k: v < alphas for k, v in p_values.items()
            }

        self.stats_per_target = stats_per_target
        return self

    def get_results_table(self):
        assert self.stats_per_target is not None, "compute_stats_per_target first"
        targets = self.target_variant_means.keys()
        results = dict()
        for variant in range(1, self.n_variants):
            df_partial_results = pd.DataFrame(
                {
                    "Control_Group_Count": [
                        self.variant_n_samples.get(self.CONTROL_LABEL, 0)
                    ]
                    * self.n_targets,
                    "Treatment_Group_Count": [self.variant_n_samples.get(variant, 0)]
                    * self.n_targets,
                    "Control_Group_Mean": [
                        self.target_variant_means[target].get(
                            self.CONTROL_LABEL, np.nan
                        )
                        for target in targets
                    ],
                    "Treatment_Group_Mean": [
                        self.target_variant_means[target].get(variant, np.nan)
                        for target in targets
                    ],
                    "Estimated_Effect_absolute": [
                        self.stats_per_target[target]["absolute_tes"].get(
                            variant, np.nan
                        )
                        for target in targets
                    ],
                    "Estimated_Effect_relative": [
                        self.stats_per_target[target]["relative_tes"].get(
                            variant, np.nan
                        )
                        for target in targets
                    ],
                    "p_value": [
                        self.stats_per_target[target]["p_values"].get(variant, 1)
                        for target in targets
                    ],
                    "is_significant": [
                        self.stats_per_target[target]["is_significant"].get(
                            variant, False
                        )
                        for target in targets
                    ],
                    "test_statistic": [
                        self.stats_per_target[target]["test_stats"].get(variant, np.nan)
                        for target in targets
                    ],
                    "Test_type": [
                        self.stats_per_target[target]["test_type"] for target in targets
                    ],
                    "CI_lower": [
                        self.stats_per_target[target]["cis_lower"].get(variant, -np.inf)
                        for target in targets
                    ],
                    "CI_upper": [
                        self.stats_per_target[target]["cis_upper"].get(variant, np.inf)
                        for target in targets
                    ],
                },
                targets,
            )
            results[variant] = df_partial_results
        results_table = pd.concat(results.values(), axis=0, keys=results.keys())
        results_table.index.names = ["Variant", "Outcome"]
        results_table["p_value"] = results_table["p_value"].round(4)
        results_table["Estimated_Effect_relative"] = results_table[
            "Estimated_Effect_relative"
        ].replace(np.inf, 0)
        return results_table


def compute_frequentist_results(
    ed: ExperimentDataset,
):
    variant_n_samples = ed.data[ed.variant].value_counts().to_dict()

    target_variant_means = ed.data.groupby(ed.variant)[ed.targets].mean().to_dict()
    target_variant_stds = ed.data.groupby(ed.variant)[ed.targets].std().to_dict()
    target_metric_types = {k: v for k, v in ed.metric_types.items() if k in ed.targets}
    n_variants = ed.n_variants
    ftr = FrequentistTestResults(
        target_variant_means=target_variant_means,
        target_variant_stds=target_variant_stds,
        variant_n_samples=variant_n_samples,
        target_metric_types=target_metric_types,
        n_variants=n_variants,
    )
    return ftr


class MulitpleHypothesisCorrection(ABC):
    def __init__(
        self,
        correct_n_variants: bool = True,
        correct_n_targets: bool = False,
        correct_n_segments: bool = False,
        n_treatments: int = 1,
        n_targets: int = 1,
        n_segments: int = 0,
        type_i_error: float = 0.05,
    ):
        self.alpha = type_i_error
        self.correct_n_variants = correct_n_variants
        self.correct_n_targets = correct_n_targets
        self.correct_n_segments = correct_n_segments
        self.segments_label = None
        self.n_treatments = n_treatments
        self.n_targets = n_targets
        self.n_segments = n_segments
        self.type_i_error = type_i_error

    @abstractmethod
    def correct_p_values(self, p_values):
        pass

    @abstractmethod
    def correct_alphas(self, p_values=None):
        pass


class NoCorrection(MulitpleHypothesisCorrection):
    def correct_p_values(self, p_values: Dict[str, float]):
        return p_values

    def correct_alphas(self, p_values=None):
        return self.type_i_error


class BonferroniCorrection(MulitpleHypothesisCorrection):
    def correct_p_values(self, p_values: Dict[str, float]):
        if self.correct_n_targets or self.correct_n_segments:
            raise NotImplementedError(
                "Correction for targets and segments not implemented yet"
            )
        elif self.correct_n_variants:
            n_hypotheses = self.n_treatments
            p_values_corrected = {
                k: min(v * n_hypotheses, 1) for k, v in p_values.items()
            }
            return p_values_corrected
        else:
            return p_values

    def correct_alphas(self, p_values: Dict[str, float] = None):
        """Corrects the alpha values for multiple hypothesis testing.

        Args:
            p_values (list, optional): List of p-values for each hypothesis.
                Defaults to None.

        Raises:
            NotImplementedError: If correction for targets and segments
            is not implemented yet.

        Returns:
            Union[Dict, float]: corrected alpha values for each hypothesis
                or a single value for all
        """
        if self.correct_n_targets or self.correct_n_segments:
            raise NotImplementedError(
                "Correction for targets and segments not implemented yet"
            )
        elif self.correct_n_variants:
            alpha_corrected = self.type_i_error / self.n_treatments
            return alpha_corrected
        else:
            return self.type_i_error


class BaseTest(ABC):
    def __init__(
        self,
        ed: ExperimentDataset,
    ):
        self.ed = ed
        # decision in ['Reject', 'Accept', 'Continue', None]
        self.decision = None
        self.fitted = False

        self._naive_estimates()

    def get_decision(self):
        pass

    def _check_fitted(self):
        if not self.fitted:
            raise Exception("Model needs to be fitted first.")
        else:
            pass

    def _naive_estimates(self):
        self.naive_counts = {target: {} for target in self.ed.targets}
        for target in self.ed.targets:
            self.naive_counts[target] = {
                label: self.outcomes_of_variant(label, target).count()
                for label in self.ed.variant_labels
            }

        self.naive_means = {target: {} for target in self.ed.targets}
        for target in self.ed.targets:
            self.naive_means[target] = {
                label: self.outcomes_of_variant(label, target).mean()
                for label in self.ed.variant_labels
            }

        self.naive_treatment_effects = {target: {} for target in self.ed.targets}

        for target in self.ed.targets:
            self.naive_treatment_effects[target] = {
                label: (
                    self.naive_means[target][label]
                    - self.naive_means[target][self.ed.control_label]
                )
                for label in set(self.ed.variant_labels) - set([self.ed.control_label])
            }

        self.naive_relative_treatment_effects = {
            target: {} for target in self.ed.targets
        }

        for target in self.ed.targets:
            self.naive_relative_treatment_effects[target] = {
                label: self._naive_relative_treatment_effects(target, label)
                for label in set(self.ed.variant_labels) - set([self.ed.control_label])
            }

    def _naive_relative_treatment_effects(self, target, label):
        try:
            relative_te = (
                self.naive_treatment_effects[target][label]
                / self.naive_means[target][self.ed.control_label]
                * 100
            )
        except ZeroDivisionError:
            relative_te = 0

        return relative_te

    def outcomes_of_variant(self, variant, target):
        # TODO: also define this function in ExperimentDataset class
        return self.ed.data.loc[(self.ed.data[self.ed.variant] == variant), target]


def t_test(sample_1, sample_2, alpha, alternative="two-sided"):
    stat, p_value, deg_freedom = ttest_ind(
        sample_1, sample_2, usevar="unequal", alternative=alternative
    )
    x_delta = sample_2.mean() - sample_1.mean()
    sd = np.sqrt(
        (sample_2.std() / np.sqrt(len(sample_2))) ** 2
        + (sample_1.std() / np.sqrt(len(sample_1))) ** 2
    )

    if alternative == "two-sided":
        conf_interval = (
            x_delta - sps.t.ppf(1 - (alpha / 2), deg_freedom) * sd,
            x_delta + sps.t.ppf(1 - (alpha / 2), deg_freedom) * sd,
        )
    return p_value, conf_interval, stat


def proportion_test(sample_1, sample_2, alpha, alternative="two-sided"):
    stat, p_value = ztest_proportions_2indep(
        sample_1.sum(),
        len(sample_1),
        sample_2.sum(),
        len(sample_2),
        alternative=alternative,
    )
    conf_interval = confint_proportions_2indep(
        sample_1.sum(),
        len(sample_1),
        sample_2.sum(),
        len(sample_2),
        alpha=alpha,
    )
    return p_value, conf_interval, stat


AVAILABLE_TESTS = {
    "binary": ["proportion_test"],
    "discrete": ["t_test"],
    "continuous": ["t_test"],
}


class FrequentistTest(BaseTest):
    """Frequentist static testing Should include test stat, confidence interval, p
    values, power,"""

    def __init__(
        self,
        ed: ExperimentDataset,
        alpha: float = 0.05,
        beta: float = 0.8,
        multitest_correction="bonferroni",
    ):
        super().__init__(ed)

        # TODO: assert that keys agree with metric types
        self.test_types = {
            "binary": "proportion_test",
            "discrete": "t_test",
            "continuous": "t_test",
        }

        self.alpha = alpha
        self.beta = beta
        # TODO: multitest_correction in [None, 'bonferroni', ...]
        self.multitest_correction = multitest_correction

    def compute(self):
        # main method to generate results

        alpha = self.alpha
        if self.multitest_correction is not None:
            self._alpha_correction()
            alpha = self.alpha_corrected

        variant_dict = {
            variant: {}
            for variant in set(self.ed.variant_labels) - set([self.ed.control_label])
        }

        self.p_values = copy.deepcopy(variant_dict)
        self.conf_intervals = copy.deepcopy(variant_dict)
        self.test_stats = copy.deepcopy(variant_dict)

        for variant in set(self.ed.variant_labels) - set([self.ed.control_label]):
            for target in self.ed.targets:
                assert (
                    self.test_types[self.ed.metric_types[target]]
                    in AVAILABLE_TESTS[self.ed.metric_types[target]]
                )

                (
                    self.p_values[variant][target],
                    self.conf_intervals[variant][target],
                    self.test_stats[variant][target],
                ) = self._run_test(
                    self.test_types[self.ed.metric_types[target]],
                    self.outcomes_of_variant(self.ed.control_label, target),
                    self.outcomes_of_variant(variant, target),
                    alpha=alpha,
                )
            self._correct_pvalues(variant)

    @staticmethod
    def _run_test(test_type, data_control, data_treat, alpha):
        if test_type == "t_test":
            p_value, conf_interval, stat = t_test(
                data_control, data_treat, alpha, alternative="two-sided"
            )
        elif test_type == "proportion_test":
            # There seems to be a statsmodels bug forcing us to swap control
            # and treat data here
            p_value, conf_interval, stat = proportion_test(
                data_treat, data_control, alpha, alternative="two-sided"
            )
        return p_value, conf_interval, stat

    def set_hypothesis_test(self, metric_type: str, hypothesis_test: str):
        # set test for a given metric,
        # e.g. set_hypothesis_test('continuous', 't_test')
        self.test_type[metric_type] = hypothesis_test

    def set_alpha(self):
        self.fitted = False
        pass

    def set_beta(self):
        self.fitted = False
        pass

    def set_data_model(self):
        # set new ExperimentDataset
        self.fitted = False
        pass

    def _compute_confidence_intervals(self, method: str = "analytical_or_bootstrap"):
        # method must be in
        # ['analytical_or_bootstrap', 'no_bootstrap', 'always_bootstrap']
        pass

    def _bootstrap_confidence_intervals(self):
        # implement scipy.stats.bootstrap
        pass

    def _alpha_correction(self, correction_level="per_treatment"):
        # correction_level in ['per_treatment', 'global']
        # If global, correction parameter is n_treatments*n_targets.
        # Else it is n_targets
        if self.multitest_correction == "bonferroni":
            n_hypotheses = (self.ed.n_variants - 1) * (
                correction_level == "per_treatment"
            ) + len(self.ed.targets) * (self.ed.n_variants - 1) * (
                correction_level == "global"
            )
            self.alpha_corrected = self.alpha / n_hypotheses
        else:
            pass

    def _correct_pvalues(self, variant, correction_level="per_treatment"):
        if self.multitest_correction == "bonferroni":
            n_hypotheses = (self.ed.n_variants - 1) * (
                correction_level == "per_treatment"
            ) + len(self.ed.targets) * (self.ed.n_variants - 1) * (
                correction_level == "global"
            )
            self.p_values[variant] = {
                target: min(self.p_values[variant][target] * n_hypotheses, 1)
                for target in self.ed.targets
            }

    def get_results_table(self):
        # control group, treatment group sizes
        # Then: one table per treatment:
        #   one row per metric:
        #       variant,
        #       control group count
        #       treatment group count
        #       control group mean (add required sample size later)
        #       treatment group mean (add required sample size later)
        #       Absolute difference
        #       Relative difference
        #       confidence interval
        #       p value
        #       is significant or not
        #       [test stat, test type] add in verbose option

        # self._check_fitted()

        results = {}
        for variant in range(1, self.ed.n_variants):
            df_result = pd.DataFrame(
                {
                    "Control_Group_Count": [
                        self.naive_counts[target][self.ed.control_label]
                        for target in self.ed.targets
                    ],
                    "Treatment_Group_Count": [
                        self.naive_counts[target][variant] for target in self.ed.targets
                    ],
                    "Control_Group_Mean": [
                        self.naive_means[target][self.ed.control_label]
                        for target in self.ed.targets
                    ],
                    "Treatment_Group_Mean": [
                        self.naive_means[target][variant] for target in self.ed.targets
                    ],
                    "Estimated_Effect_absolute": [
                        self.naive_treatment_effects[target][variant]
                        for target in self.ed.targets
                    ],
                    "Estimated_Effect_relative": [
                        self.naive_relative_treatment_effects[target][variant]
                        for target in self.ed.targets
                    ],
                    "Confidence_Interval": self.conf_intervals[variant].values(),
                    "p_value": self.p_values[variant].values(),
                    "is_significant": [
                        1 if self.p_values[variant][target] < self.alpha else 0
                        for target in self.ed.targets
                    ],
                    "test_statistic": self.test_stats[variant].values(),
                    "Test_type": [
                        self.test_types[self.ed.metric_types[target]]
                        for target in self.ed.targets
                    ],
                },
                self.ed.targets,
            )
            df_result[["CI_lower", "CI_upper"]] = pd.DataFrame(
                df_result["Confidence_Interval"].tolist(), index=df_result.index
            )
            del df_result["Confidence_Interval"]

            results[variant] = df_result
            results_table = pd.concat(results.values(), axis=0, keys=results.keys())
            results_table.index.names = ["Variant", "Outcome"]
        results_table["p_value"] = results_table["p_value"].round(4)
        results_table["Estimated_Effect_relative"] = results_table[
            "Estimated_Effect_relative"
        ].replace(np.inf, 0)
        return results_table


def cuped(ed: ExperimentDataset, has_correction: bool, alpha: float):
    """Applies the CUPED method to estimate treatment effects in experiment. Serves as a
    wrapper to variance reduction methods.

    Args:
        ExperimentDataset (ExperimentDataset): An ExperimentDataset object
            containing the data to be analyzed.
        has_correction (bool): A boolean indicating whether to apply a
            multitest correction.
            If True, the Bonferroni correction is applied.
        alpha (float): The significance level for hypothesis testing.

    Returns:
        final_df (DataFrame): A pandas DataFrame containing the estimated
            treatment effects, confidence intervals,
            p-values, and significance levels for each target and treatment combination.
    """
    # check if multiple covariates are to be used
    multivariate = (
        isinstance(ed.pre_experiment_cols, list) and len(ed.pre_experiment_cols) > 1
    )

    # create dummy variables for variants
    dummified_data = pd.get_dummies(ed.data, columns=[ed.variant], drop_first=True)

    # create list of treatments
    treatments = [i for i in range(1, ed.n_variants)]

    # multitest correction NOTE: copied straight from FrequentistTest._alpha_correction
    # TODO: remove duplication, integrate with MultipleHypothesisCorrection
    multitest_correction = (
        "bonferroni" if ed.n_variants > 2 and has_correction == "Yes" else None
    )
    if multitest_correction is not None:
        if multitest_correction == "bonferroni":
            correction_level = "per_treatment"  # NOTE: hardcoded for the time being
            n_hypotheses = (ed.n_variants - 1) * (
                correction_level == "per_treatment"
            ) + len(ed.targets) * (ed.n_variants - 1) * (correction_level == "global")
            alpha_corrected = alpha / n_hypotheses
            alpha = alpha_corrected
        else:
            raise NotImplementedError
    else:
        pass

    # initiate estimator
    if multivariate:
        estimator = MultivariateCUPED()

    else:
        estimator = CUPED()

    result_dfs = []

    # run estimator for each target and treatment
    for target in ed.targets:
        for treatment in treatments:
            estimator = estimator.fit(
                data=dummified_data,
                treatment_column=ed.variant + "_" + str(treatment),
                target_column=target,
                covariate_column=ed.pre_experiment_cols[0],
                covariate_columns=ed.pre_experiment_cols,
            )

            result_df = pd.DataFrame(
                {
                    "Variant": [treatment],
                    "Outcome": [target],
                    "Estimated_Effect_absolute": [estimator.estimate],
                    "CI_lower": [
                        estimator.regression_results.conf_int(alpha=alpha, cols=None)[
                            0
                        ][1]
                    ],
                    "CI_upper": [
                        estimator.regression_results.conf_int(alpha=alpha, cols=None)[
                            1
                        ][1]
                    ],
                    "p_value": [estimator.p_value],
                    "is_significant": [estimator.p_value < alpha],
                }
                # ed.targets,
            )
            result_dfs.append(result_df)

    # concatenate results
    final_df = pd.concat(result_dfs)
    final_df = final_df.set_index(["Variant", "Outcome"])

    return final_df


def run_cuped(ed: ExperimentDataset):
    """Applies the CUPED method to estimate treatment effects in experiment.
    Serves as a wrapper to variance reduction methods.

    Args:
        ExperimentDataset (ExperimentDataset): An ExperimentDataset object
            containing the data to be analyzed.

    Returns:
        output (CUPEDOutput): A results object with all fitted estimators
        that can then be assembled into a results table.
    """
    # check if multiple covariates are to be used
    multivariate = (
        isinstance(ed.pre_experiment_cols, list) and len(ed.pre_experiment_cols) > 1
    )

    # create dummy variables for variants
    dummified_data = pd.get_dummies(ed.data, columns=[ed.variant], drop_first=True)

    # create list of treatments
    treatments = [i for i in range(1, ed.n_variants)]

    estimators = dict()

    # run estimator for each target and treatment
    for target in ed.targets:
        estimators[target] = dict()
        for treatment in treatments:
            # initiate estimator
            if multivariate:
                estimator = MultivariateCUPED()

            else:
                estimator = CUPED()

            estimators[target][treatment] = estimator.fit(
                data=dummified_data,
                treatment_column=ed.variant + "_" + str(treatment),
                target_column=target,
                covariate_column=ed.pre_experiment_cols[0],
                covariate_columns=ed.pre_experiment_cols,
            )

    return CUPEDoutput(estimators=estimators, n_variants=ed.n_variants)


@dataclass
class CUPEDoutput:
    estimators: Dict[str, Dict[str, float]]
    n_variants: int

    def results_table(self, alpha: float = 0.05, has_correction: bool = False):
        # multitest correction
        # NOTE: copied straight from FrequentistTest._alpha_correction
        # TODO: remove duplication, integrate with MultipleHypothesisCorrection
        multitest_correction = (
            "bonferroni" if self.n_variants > 2 and has_correction == "Yes" else None
        )
        if multitest_correction is not None:
            if multitest_correction == "bonferroni":
                correction_level = "per_treatment"  # NOTE: hardcoded for now
                n_hypotheses = (self.n_variants - 1) * (
                    correction_level == "per_treatment"
                ) + len(self.targets) * (self.n_variants - 1) * (
                    correction_level == "global"
                )
                alpha_corrected = alpha / n_hypotheses
                alpha = alpha_corrected
            else:
                raise NotImplementedError
        else:
            pass

        result_dfs = []
        for target in self.estimators.keys():
            for treatment in range(1, self.n_variants):
                estimator = self.estimators[target][treatment]
                result_df = pd.DataFrame(
                    {
                        "Variant": [treatment],
                        "Outcome": [target],
                        "Estimated_Effect_absolute": [estimator.estimate],
                        "CI_lower": [
                            estimator.regression_results.conf_int(
                                alpha=alpha, cols=None
                            )[0][1]
                        ],
                        "CI_upper": [
                            estimator.regression_results.conf_int(
                                alpha=alpha, cols=None
                            )[1][1]
                        ],
                        "p_value": [estimator.p_value],
                        "is_significant": [estimator.p_value < alpha],
                    }
                    # ed.targets,
                )
            result_dfs.append(result_df)

        # concatenate results
        final_df = pd.concat(result_dfs)
        final_df = final_df.set_index(["Variant", "Outcome"])
        return final_df



from numpyro.distributions import (
    ZeroInflatedPoisson,
    BernoulliProbs,
    LogNormal,
    Gamma,
    Uniform,
    Beta,
    Normal,
)
from tw_experimentation.constants import MetricType


# AVAILABLE_LIKELIHOOD_MODELS = {
#     "binary": [Bernoulli],
#     "continuous": [ZeroInflatedLogNormal, Normal, Exponential],
#     "discrete": [ZeroInflatedPoisson],
# }


# LIKELIHOOD_PRIORS = {
#     "Bernoulli": {"probs": Beta},
#     "ZeroInflatedLogNormal": {"loc": LogNormal, "scale": Gamma, "gate": Uniform},
#     "Normal": {"loc": LogNormal, "scale": Gamma},
#     "ZeroInflatedPoisson": {"rate": Gamma, "gate": Beta},
#     "Exponential": {"rate": Gamma},
# }

# EXPECTED_VALUE_PER_LIKELIHOOD = {}

# PRIOS_PARAMS = {
#     "LogNormal": {"args": {}, "kwargs": {"loc": 0, "scale": 1}},
#     "Gamma": {"args": (1,), "kwargs": {"rate": 0.5}},
#     "Uniform": {"args": (), "kwargs": {"low": 0, "high": 1}},
#     "Beta": {"args": (2, 2), "kwargs": {}},
#     "Gamma": {"args": (1,), "kwargs": {"rate": 0.5}},
#     "Beta": {"args": (2, 2), "kwargs": {}},
# }

# class Likelihood(ABCMeta):
#     @property
#     @abstractmethod
#     def metrictype(self):
#         raise NotImplementedError()

#     @property
#     @abstractmethod
#     def auxiliary_zero_inflation(self):
#         raise NotImplementedError()


class BernoulliLikelihood(BernoulliProbs):
    def metrictype(self):
        return MetricType.BINARY

    @property
    def auxiliary_zero_inflation(self):
        return False

    @property
    def aux_dist(self):
        if not self.auxiliary_zero_inflation:
            return None
        else:
            return Uniform(low=0, high=1)


class ZeroInflatedLogNormalLikelihood(LogNormal):
    def metrictype(self):
        return MetricType.CONTINUOUS

    @property
    def auxiliary_zero_inflation(self):
        return True

    def aux_dist(self):
        if not self.auxiliary_zero_inflation:
            return None
        else:
            return Uniform


class ZeroInflatedPoissonLikelihood(ZeroInflatedPoisson):
    def metrictype(self):
        return MetricType.DISCRETE

    @property
    def auxiliary_zero_inflation(self):
        return False

    @property
    def aux_dist(self):
        if not self.auxiliary_zero_inflation:
            return None
        else:
            return Uniform


DEFAULT_LIKELIHOOD_MODELS = {
    "binary": BernoulliLikelihood,
    "continuous": ZeroInflatedLogNormalLikelihood,
    "discrete": ZeroInflatedPoissonLikelihood,
}


DEFAULT_PRIOR_MODELS = {
    "binary": {"probs": Beta},
    "continuous": {
        "loc": Normal,
        "scale": Gamma,
        # "gate": Uniform
    },
    "discrete": {"rate": Gamma, "gate": Beta},
}


DEFAULT_PRIOR_MODEL_PARAMS = {
    "binary": {"probs": {"args": (2, 2), "kwargs": {}}},
    "continuous": {
        "loc": {"args": {}, "kwargs": {"loc": 0, "scale": 0.25}},
        "scale": {"args": (0.5,), "kwargs": {"rate": 0.5}},
        # "gate": {"args": (), "kwargs": {"low": 0, "high": 1}},
    },
    "discrete": {
        "rate": {"args": (1,), "kwargs": {"rate": 0.5}},
        "gate": {"args": (2, 2), "kwargs": {}},
    },
}


# class BernoulliLikelihood(Likelihood, Bernoulli):
#     def metrictype(self):
#         return MetricType.BINARY

#     def param_names(self):
#         return {"probs": BetaPrior()}

#     def expected_value(self):
#         return "probs"


# class ZeroInflatedPoissonLikelihood(Likelihood, ZeroInflatedPoisson):
#     def metrictype(self):
#         return MetricType.DISCRETE


# class ZeroInflatedLogNormalLikelihood(Likelihood, ZeroInflatedLogNormal):
#     def metrictype(self):
#         return MetricType.CONTINUOUS


# class Prior:
#     @property
#     def args(self):
#         raise NotImplementedError()

#     @property
#     def kwargs(self):
#         raise NotImplementedError()


# class BetaPrior(Prior, Beta):
#     def args(self):
#         return (2, 2)

#     def kwargs(self):
#         return {}



import numpy as np


from tw_experimentation.utils import ExperimentDataset, highlight
from tw_experimentation.statistical_tests import FrequentistTest

from tw_experimentation.setuper import Setuper, effect_size_to_uplift
from tw_experimentation.segmentation_frequentist import Segmentation

import plotly.graph_objects as go

import ipywidgets as widgets
from ipywidgets import interact

from IPython.display import display

METRIC_TYPE_OPTIONS = [
    "binary",
    # "discrete",
    "continuous",
]
DAY_COL = "day"


class MonitoringInterface:
    def __init__(self, ed) -> None:
        self.ed = ed
        # self.monitor = Monitoring(self.ed)
        # self.monitor._plot_sample_ratio_mismatch()
        # self.monitor.target_monitoring()
        """
        self.monitor_metric_wt = widgets.SelectMultiple(
            options=extra_cols, description="Outcome"
        )

        @widgets.interact_manual(metrics=self.monitor_metric_wt)
        def plot_extra_metric(metrics):
            metrics = list(metrics)
            self.monitor.sanity_metric_monitoring(metrics)

        self.target_wt = widgets.Select(options=self.ed.targets, description="Outcome")

        self.segment_wt = widgets.Select(options=extra_cols, description="Segment")

        @widgets.interact_manual(target=self.target_wt, segment=self.segment_wt)
        def plot_segment(target, segment):
            self.monitor.segment_monitoring(target, segment)
        """


class SegmentationInterface:
    def __init__(self, ed) -> None:
        self.ed = ed

        extra_cols = (
            set(self.ed.data.columns)
            - set(self.ed.targets)
            - set([self.ed.date])
            - set([self.ed.variant])
            - set([DAY_COL])
        )

        self.segment_choice_wt = widgets.Dropdown(
            options=extra_cols, description="Segment"
        )
        title_wt = widgets.HTML(value="<h2><b>Evaluation: Segmentation</b></h2>")
        display(title_wt)

        @widgets.interact_manual(category=self.segment_choice_wt)
        def manual_segmentation(category):
            s = Segmentation(self.ed)
            df_summary = s.frequentist_by_segment(category)
            display(
                df_summary.style.apply(highlight, axis=1)
                .set_table_attributes("style='display:inline'")
                .bar(subset=["Estimated_Effect_relative"], color="grey")
                .format(precision=3)
            )


class SampleSizeInterface:
    def __init__(self) -> None:
        pass

    def classical_test(self):
        pass
        # treatment share
        # if continuous: sd
        # if binary:
        # baseline conversion
        # uplift relative or absolute?

        self.alpha_widget = widgets.FloatSlider(
            value=0.05,
            min=0,
            max=1.0,
            step=0.01,
            description="Alpha (Type-I error):",
            continuous_update=True,
            orientation="horizontal",
            readout=True,
            readout_format=".2f",
            style=dict(description_width="initial"),
        )

        self.beta_widget = widgets.FloatSlider(
            value=0.2,
            min=0,
            max=1.0,
            step=0.01,
            description="Beta (Type-II error):",
            continuous_update=True,
            orientation="horizontal",
            readout=True,
            readout_format=".2f",
            style=dict(description_width="initial"),
        )
        self.type_widget = widgets.ToggleButtons(
            options=METRIC_TYPE_OPTIONS,
            description="Key outcome metric type:",
            disabled=False,
            style=dict(description_width="initial"),
        )
        self.tshare_widget = widgets.FloatSlider(
            value=0.5,
            min=0,
            max=1.0,
            step=0.01,
            description="Treatment Share:",
            disabled=False,
            continuous_update=True,
            orientation="horizontal",
            readout=True,
            readout_format=".2f",
            style=dict(description_width="initial"),
        )
        # binary outcome specifications
        self.bc_widget = widgets.FloatSlider(
            value=0.5,
            min=0,
            max=1.0,
            step=0.001,
            description="Baseline Conversion:",
            disabled=False,
            continuous_update=True,
            orientation="horizontal",
            readout=True,
            readout_format=".3f",
            style=dict(description_width="initial"),
        )

        self.et_widget = widgets.ToggleButtons(
            options=["absolute", "relative"],
            description="Effect type:",
            disabled=False,
            style=dict(description_width="initial"),
        )
        self.es_widget = widgets.FloatText(
            value=0.05,
            description="Uplift to detect:",
            disabled=False,
            readout_format=".3f",
            style=dict(description_width="initial"),
        )

        # continuous outcome specifications
        self.sd_widget = widgets.FloatText(
            value=1,
            description="standard deviation:",
            disabled=False,
            style=dict(description_width="initial"),
        )

        self.mean_widget = widgets.FloatText(
            value=100,
            description="baseline mean:",
            disabled=False,
            style=dict(description_width="initial"),
        )

        self.type_widget.observe(self._continuous_binary_switch, names="value")

        self.sd_widget.layout.visibility = "hidden"
        self.mean_widget.layout.visibility = "hidden"

        title_wt = widgets.HTML(
            value="<h1><b>Sample Size Calculator</b></h1>",
        )
        display(title_wt)
        """        @widgets.interactive(
            alpha=self.alpha_widget,
            beta=self.beta_widget,
            tshare=self.tshare_widget,
            effect_type=self.et_widget,
            effect_size=self.es_widget,
            metric_type=self.type_widget,
            baseline_conversion=self.bc_widget,
            sd=self.sd_widget,
            mean=self.mean_widget,
        )
        """

        def tester(
            alpha,
            beta,
            tshare,
            effect_type,
            effect_size,
            metric_type,
            baseline_conversion,
            mean,
            sd,
        ):
            if metric_type == "binary":
                sd = np.sqrt(baseline_conversion * (1 - baseline_conversion))
                setup = Setuper.from_uplift(
                    alpha=alpha,
                    beta=beta,
                    uplift=effect_size,
                    sd=sd,
                    mean=baseline_conversion,
                    relation=effect_type,
                    treatment_share=tshare,
                )
                sample_size = setup.sample_size_two_sample_proportion_z_test()
                max_sample_size = 2 * sample_size["Total Sample Size"]
                x_sample_size = np.linspace(
                    200 if max_sample_size > 500 else 20, max_sample_size, num=100
                )

                def uplift_map(x):
                    effect_size_to_uplift(
                        setup.effect_size_two_sample_z_test(x),
                        baseline_conversion,
                        sd,
                        relation=effect_type,
                    )

                uplift = np.array(list(map(uplift_map, x_sample_size)))

                fig = go.FigureWidget(data=go.Scatter(x=x_sample_size, y=uplift))
                fig.add_vline(
                    x=sample_size["Total Sample Size"],
                    line_width=3,
                    line_dash="dash",
                    line_color="green",
                )
                fig.update_xaxes(title_text="Sample Size")
                fig.update_yaxes(title_text="Minimum Detectable Uplift")

            elif metric_type in ["discrete", "continuous"]:
                setup = Setuper.from_uplift(
                    alpha=alpha,
                    beta=beta,
                    uplift=effect_size,
                    sd=sd,
                    mean=mean,
                    relation=effect_type,
                    treatment_share=tshare,
                )
                sample_size = setup.sample_size_t_test()
                max_sample_size = 2 * sample_size["Total Sample Size"]
                x_sample_size = np.linspace(
                    200 if max_sample_size > 500 else 20, max_sample_size, num=100
                )

                def uplift_map(x):
                    effect_size_to_uplift(
                        setup.effect_size_t_test(x),
                        mean,
                        sd,
                        relation=effect_type,
                    )

                uplift = np.array(list(map(uplift_map, x_sample_size)))
                fig = go.FigureWidget(data=go.Scatter(x=x_sample_size, y=uplift))
                fig.add_vline(
                    x=sample_size["Total Sample Size"],
                    line_width=3,
                    line_dash="dash",
                    line_color="green",
                )
                fig.update_xaxes(title_text="Sample Size")
                fig.update_yaxes(title_text="Minimum Detectable Uplift")
            sample_size_total = (
                sample_size["Treatment Sample Size"]
                + sample_size["Control Sample Size"]
            )
            widget_val = (
                "<h2>Total Sample Size:"
                f" {sample_size_total}"
                "</h2><h3>Treatment"
                f" Sample Size: {sample_size['Treatment Sample Size']}<br> Control"
                f" Sample Size: {sample_size['Control Sample Size']}</h3>"
            )
            result_wt = widgets.HTML(value=widget_val)
            display(result_wt)
            display(fig)

        wt = widgets.interactive(
            tester,
            alpha=self.alpha_widget,
            beta=self.beta_widget,
            tshare=self.tshare_widget,
            effect_type=self.et_widget,
            effect_size=self.es_widget,
            metric_type=self.type_widget,
            baseline_conversion=self.bc_widget,
            sd=self.sd_widget,
            mean=self.mean_widget,
        )
        display(wt)

    def _continuous_binary_switch(self, change):
        if self.type_widget.value == "binary":
            self.bc_widget.layout.visibility = "visible"
            # self.et_widget.layout.visibility = "visible"
            self.sd_widget.layout.visibility = "hidden"
            self.mean_widget.layout.visibility = "hidden"
        else:
            self.bc_widget.layout.visibility = "hidden"
            # self.et_widget.layout.visibility = "hidden"
            self.sd_widget.layout.visibility = "visible"
            self.mean_widget.layout.visibility = "visible"


class DataWidget:
    def __init__(self, df) -> None:
        self.df = df

        title_wt = widgets.HTML(
            value=(
                "<h2><b>Defining the Data:</b></h2> <h3>Please select the relevant"
                " columns</h3>"
            ),
        )
        display(title_wt)

        self.target_widget = widgets.SelectMultiple(
            options=self.df.columns,
            description="Outcomes",
            style=dict(description_width="initial"),
        )
        display(self.target_widget)

        self._variant_col_name()
        self._date_label()

        self.pre_exp_cols = widgets.SelectMultiple(
            options=self.df.columns,
            description="Pre-Experiment Columns",
            style=dict(description_width="initial"),
        )
        display(self.pre_exp_cols)

        self.mts = {}
        for col in self.df.columns:
            self.mts[col] = widgets.ToggleButtons(
                options=METRIC_TYPE_OPTIONS,
                description=f"Metric Type of {col}",
                disabled=False,
                style=dict(description_width="initial"),
            )
            display(self.mts[col])
            self.mts[col].layout.visibility = "hidden"
            self.mts[col].layout.display = "none"

        self.target_widget.observe(self._target_change, names="value")

    def create_data_model(self):
        dm_button = widgets.Button(description="Create data model!")

        dm_button.on_click(self._create_experiment_dataset)
        display(dm_button)

    def _target_change(self, change):
        for target in set(change.old):
            self.mts[target].layout.visibility = "hidden"

        for target in set(change.new):
            self.mts[target].layout.visibility = "visible"
            self.mts[target].layout.display = "block"

    def _variant_col_name(self):
        self.variant_col = widgets.Dropdown(
            options=self.df.columns,
            description="Name of Column with Variants",
            style=dict(description_width="initial"),
        )
        display(self.variant_col)

    def _date_label(self):
        self.timestamp_col = widgets.Dropdown(
            options=self.df.columns,
            description="Name of Column with Timestamps",
            style=dict(description_width="initial"),
        )
        display(self.timestamp_col)

    def _create_experiment_dataset(self, b):
        self.ed = ExperimentDataset(
            data=self.df,
            variant=self.variant_col.value,
            targets=self.target_widget.value,
            date=self.timestamp_col.value,
            pre_experiment_cols=self.pre_exp_cols.value,
            n_variants=self.df[self.variant_col.value].nunique(),
            metric_types={
                target: self.mts[target].value for target in self.target_widget.value
            },
        )
        self.ed.preprocess_dataset()


class FrequentistEvaluation:
    def __init__(self, ed) -> None:
        self.ed = ed

    def start(self):
        title_wt_freq = widgets.HTML(value="<h2><b>Test Evaluation</b></h2>")
        display(title_wt_freq)

        if self.ed.n_variants > 2:
            interact(
                self._frequentist_process,
                has_correction=widgets.Dropdown(
                    options=["No", "Yes"],
                    value="Yes",
                    description="multi-test correction:",
                    disabled=False,
                    style=dict(description_width="initial"),
                ),
            )

        else:
            self._frequentist_process(False)

    def _frequentist_process(self, has_correction):
        multitest_correction = None
        if has_correction == "Yes":
            multitest_correction = "bonferroni"
        ft = FrequentistTest(ed=self.ed, multitest_correction=multitest_correction)
        ft.compute()

        self.result_df = ft.get_results_table()
        display(
            self.result_df.style.apply(highlight, axis=1)
            .set_table_attributes("style='display:inline'")
            .bar(subset=["Estimated_Effect_relative"], color="grey")
            .format(precision=3)
        )



"""Tool for testing ab experiments using different approaches for several metrics.

Code contains Tester class with methods for specific tests. Support only pandas
DataFrames now.
"""

import pandas as pd
import numpy as np
from statsmodels import stats as sms
from scipy import stats as scipystats

from sklearn.utils import resample

from typing import List, Union

from SequentialProbabilityRatioTTest import (
    SequentialProbabilityRatioTTest,
    SequentialProbabilityRatioBinaryTest,
)


class BinaryTest:
    def __init__(
        self,
        df: Union[pd.DataFrame, None] = None,
        user_id_column: Union[str, None] = None,
        treatment: Union[str, None] = None,
        metric: Union[str, None] = None,
        method: Union[str, None] = None,
        action_date: Union[str, None] = None,
        cohen_d: Union[float, None] = 0.2,
    ):
        self.df = df
        self.user_id_column = user_id_column
        self.treatment = treatment
        self.metric = metric
        self.method = method
        self.action_date = action_date
        self.cohen_d = cohen_d
        self.ct = self.contingency_table()
        self.config_tests = {
            "Proportion": "static_proportion_test",
            "Fisher": "fishers_exact_test",
            "Chi": "chi_squared_test",
            "Dynamic": "dynamic_test",
        }

    def contingency_table(self):
        """Output contingency table of outcome for treatment and control :return: 2x2
        DataFrame."""
        ct = pd.crosstab(self.df[self.treatment], self.df[self.metric])
        idx = pd.Index(["Control", "Treatment"], name=None)
        ct.index = idx
        return ct

    def effect_estimate(self):
        return (self.ct.iloc[1, 1] / self.ct.iloc[1, :].sum()) - (
            self.ct.iloc[0, 1] / self.ct.iloc[0, :].sum()
        )

    def run(self):
        return eval("self." + self.config_tests[self.method] + "()")

    def static_proportion_test(self, direction="two-sided"):
        ct = self.ct
        result = sms.proportion.test_proportions_2indep(
            count1=ct.iloc[0, 1],
            nobs1=ct.iloc[0].sum(),
            count2=ct.iloc[1, 1],
            nobs2=ct.iloc[1].sum(),
            alternative=direction,
        )
        return {
            "statistic": result.statistic,
            "pvalue": result.pvalue,
            "variance": result.variance,
            "effect_estimate": self.effect_estimate(),
            "Decision": None,
            "further_results": {
                "contingency table": self.ct,
                "odds_ratio": result.odds_ratio,
                "ratio": result.ratio,
            },
            "plots": None,
        }

    def fishers_exact_test(self, direction="two-sided"):
        """Fisher's exact test :param df: DataFrame :param treatment: str treatment name
        :param metric: str outcome name :param direction:

        Options:
                ‘two-sided’: the odds ratio of the underlying population is not one
                ‘less’: the odds ratio of the underlying population is less than one
                ‘greater’: the odds ratio of the underlying population is greater than 1
        :return: SignificanceResult
                    statistic: float
                    pvalue: float
        """
        ct = self.ct
        result = scipystats.fisher_exact(ct, alternative=direction)
        return {
            "statistic": result[0],
            "pvalue": result[1],
            "variance": None,
            "effect_estimate": self.effect_estimate(),
            "Decision": None,
            "further_results": None,
            "plots": None,
        }

    def chi_squared_test(self):
        """
        :return:
            chi2stat: float
            p-value: float
        """
        ct = self.ct
        result = scipystats.chi2_contingency(ct)[:3]
        return {
            "statistic": result[0],
            "pvalue": result[1],
            "further_results": {"degrees_of_freedom": result[2]},
        }

    def dynamic_test(self):
        return SequentialProbabilityRatioBinaryTest(
            df=self.df,
            treatment=self.treatment,
            outcome=self.metric,
            action_date=self.action_date,
            cohen_d=self.cohen_d,
        ).run_test()


class ContinuousTest:
    def __init__(
        self,
        df: Union[pd.DataFrame, None] = None,
        user_id_column: Union[str, None] = None,
        treatment: Union[str, None] = None,
        metric: Union[str, None] = None,
        method: Union[str, None] = None,
        action_date: Union[str, None] = None,
        cohen_d: Union[float, None] = 0.2,
    ):
        self.df = df
        self.user_id_column = user_id_column
        self.treatment = treatment
        self.metric = metric
        self.method = method
        self.action_date = action_date
        self.cohen_d = cohen_d
        self.config_tests = {
            "Mann-Whitney": "mann_whitney_u_test",
            "Ttest": "ttest",
            "Welch_Ttest": "welch_ttest",
            "Dynamic": "dynamic_test",
        }

    def effect_estimate(self):
        return (
            self.df[self.metric].where(self.df[self.treatment] == 1).dropna().mean()
            - self.df[self.metric].where(self.df[self.treatment] == 0).dropna().mean()
        )

    def run(self):
        return eval("self." + self.config_tests[self.method] + "()")

    def mann_whitney_u_test(self):
        """Mann-Whitney-U test.

        To be used when sample sized is expected to be skewed / not normally distributed
        :return:
            statistic: float
            pvalue: float
        """
        result = scipystats.mannwhitneyu(
            self.df[self.metric].where(self.df[self.treatment] == 0).dropna(),
            self.df[self.metric].where(self.df[self.treatment] == 1).dropna(),
        )

        return {
            "statistic": result[0],
            "pvalue": result[1],
            "variance": None,
            "effect_estimate": self.effect_estimate(),
            "Decision": None,
            "further_results": None,
            "plots": None,
        }

    def ttest(self):
        result = sms.weightstats.ttest_ind(
            self.df[self.metric].where(self.df[self.treatment] == 0).dropna(),
            self.df[self.metric].where(self.df[self.treatment] == 1).dropna(),
        )
        return {
            "statistic": result[0],
            "pvalue": result[1],
            "variance": None,
            "effect_estimate": self.effect_estimate(),
            "Decision": None,
            "further_results": {"degrees_of_freedom": result[2]},
            "plots": None,
        }

    def welch_ttest(self):
        result = sms.weightstats.ttest_ind(
            self.df[self.metric].where(self.df[self.treatment] == 0).dropna(),
            self.df[self.metric].where(self.df[self.treatment] == 1).dropna(),
            usevar="unequal",
        )
        return {
            "statistic": result[0],
            "pvalue": result[1],
            "variance": None,
            "effect_estimate": self.effect_estimate(),
            "Decision": None,
            "further_results": {"degrees_of_freedom": result[2]},
            "plots": None,
        }

    def dynamic_test(self):
        return SequentialProbabilityRatioTTest(
            df=self.df,
            treatment=self.treatment,
            outcome=self.metric,
            action_date=self.action_date,
            cohen_d=self.cohen_d,
        ).run_test()


class ConfidenceIntervals:
    def __init__(
        self,
        df: Union[pd.DataFrame, None] = None,
        user_id_column: Union[str, None] = None,
        treatment: Union[str, None] = None,
        action_date: Union[str, None] = None,
        metric: Union[str, None] = None,
        metric_type: Union[str, None] = "Binary",
        method: Union[str, None] = None,
    ):
        self.df = df
        self.user_id_column = user_id_column
        self.treatment = treatment
        self.action_date = action_date
        self.metric = metric
        self.metric_type = metric_type
        self.method = method
        self.distribution = []

    def bootstrap_confidence_interval(self, alpha=0.05, num_iterations=2000):
        samples = self.bootstrap_samples(n_samples=num_iterations)
        if self.metric_type == "Binary":
            stat_val = BinaryTest(
                df=self.df,
                user_id_column=self.user_id_column,
                treatment=self.treatment,
                metric=self.metric,
                method=self.method,
                action_date=self.action_date,
            ).run()["effect_estimate"]
            distribution = []
            for j in range(num_iterations):
                distribution.append(
                    BinaryTest(
                        df=samples[j],
                        user_id_column=self.user_id_column,
                        treatment=self.treatment,
                        metric=self.metric,
                        method=self.method,
                        action_date=self.action_date,
                    ).run()["effect_estimate"]
                )

        elif self.metric_type == "Continuous":
            stat_val = ContinuousTest(
                df=self.df,
                user_id_column=self.user_id_column,
                treatment=self.treatment,
                metric=self.metric,
                method=self.method,
                action_date=self.action_date,
            ).run()["effect_estimate"]
            distribution = []
            for j in range(num_iterations):
                distribution.append(
                    ContinuousTest(
                        df=samples[j],
                        user_id_column=self.user_id_column,
                        treatment=self.treatment,
                        metric=self.metric,
                        method=self.method,
                        action_date=self.action_date,
                    ).run()["effect_estimate"]
                )

        self.distribution = distribution

        c_low = 2 * stat_val - np.percentile(distribution, 100 * (1 - alpha / 2.0))
        c_high = 2 * stat_val - np.percentile(distribution, 100 * (alpha / 2.0))

        return (c_low, c_high)

    def bootstrap_samples(self, n_samples=10, n_elements=2000):
        samples = []
        for _ in range(n_samples):
            sample = resample(self.df, n_samples=n_elements)
            samples.append(sample)
        return samples

    def bootstrap_treatment_control_samples(self, n_samples=10, n_elements=2000):
        treatment_samples = []
        control_samples = []
        for _ in range(n_samples):
            treatment_sample = resample(
                self.df[self.df[self.treatment] == 1], n_samples=n_elements
            )
            treatment_samples.append(treatment_sample)

            control_sample = resample(
                self.df[self.df[self.treatment] == 0], n_samples=n_elements
            )
            control_samples.append(control_sample)

            samples = treatment_samples + control_samples

        return samples


class Tester:
    """Tool for analysing AB tests.

    Result includes:
        - Statistical result
    """

    def __init__(
        self,
        df: Union[pd.DataFrame, None] = None,
        user_id_column: Union[str, None] = None,
        treatment: Union[str, None] = None,
        action_date: Union[str, None] = None,
        binary_metrics: Union[List[str], None] = None,
        continuous_metrics: Union[List[str], None] = None,
        customer_features: Union[List[str], None] = None,
    ):
        """Tester constructor."""
        self.df = df
        self.user_id_column = user_id_column
        self.treatment = treatment
        self.action_date = action_date
        self.binary_metrics = binary_metrics
        self.continuous_metrics = continuous_metrics
        self.customer_features = customer_features

    def data_info(self) -> int:
        return len(self.df)

    def ab_test(
        self,
        method_binary=None,
        method_continuous=None,
        cohen_d=None,
        direction="two-sided",
        alpha: float = 0.05,
        mult_hyp_correction=True,
    ):
        binary_test_results = (
            dict(
                [
                    (
                        metric,
                        BinaryTest(
                            df=self.df,
                            user_id_column=self.user_id_column,
                            treatment=self.treatment,
                            metric=metric,
                            method=method_binary,
                            action_date=self.action_date,
                            cohen_d=cohen_d,
                        ).run(),
                    )
                    for metric in self.binary_metrics
                ]
            )
            if self.binary_metrics is not None
            else {}
        )

        continuous_test_results = (
            dict(
                [
                    (
                        metric,
                        ContinuousTest(
                            df=self.df,
                            user_id_column=self.user_id_column,
                            treatment=self.treatment,
                            metric=metric,
                            method=method_continuous,
                            action_date=self.action_date,
                            cohen_d=cohen_d,
                        ).run(),
                    )
                    for metric in self.continuous_metrics
                ]
            )
            if self.continuous_metrics is not None
            else {}
        )

        # multiple testing correction
        #  bonferroni method
        if mult_hyp_correction:
            n_hypotheses = 0
            if self.binary_metrics is not None:
                n_hypotheses += len(self.binary_metrics)
            if self.continuous_metrics is not None:
                n_hypotheses += +len(self.continuous_metrics)

        alpha = alpha / n_hypotheses

        def decision_and_bootstrap(
            results: Union[str, None] = None,
            metrics: Union[list, None] = None,
            metric_type: Union[str, None] = None,
            method: Union[str, None] = None,
        ):
            if metrics is not None:
                for metric in metrics:
                    if (
                        "effect_estimate" in results[metric]
                        and results[metric]["effect_estimate"] is not None
                    ):
                        results[metric][
                            "confidence_interval_bootstrapped"
                        ] = ConfidenceIntervals(
                            df=self.df,
                            user_id_column=self.user_id_column,
                            treatment=self.treatment,
                            metric=metric,
                            metric_type=metric_type,
                            method=method,
                            action_date=self.action_date,
                        ).bootstrap_confidence_interval(
                            alpha=alpha
                        )
                    if (
                        "pvalue" in results[metric]
                        and results[metric]["pvalue"] is not None
                    ):
                        if results[metric]["pvalue"] < alpha:
                            results[metric]["Decision"] = "Reject H0"
                        else:
                            results[metric]["Decision"] = "Accept H0"

        decision_and_bootstrap(
            binary_test_results, self.binary_metrics, "Binary", method_binary
        )
        decision_and_bootstrap(
            continuous_test_results,
            self.continuous_metrics,
            "Continuous",
            method_continuous,
        )

        if self.binary_metrics is not None:
            for metric in self.binary_metrics:
                if (
                    "effect_estimate" in binary_test_results[metric]
                    and binary_test_results[metric]["effect_estimate"] is not None
                ):
                    binary_test_results[metric][
                        "confidence_interval_bootstrapped"
                    ] = ConfidenceIntervals(
                        df=self.df,
                        user_id_column=self.user_id_column,
                        treatment=self.treatment,
                        metric=metric,
                        metric_type="Binary",
                        method=method_binary,
                        action_date=self.action_date,
                    ).bootstrap_confidence_interval(
                        alpha=alpha
                    )
                if (
                    "pvalue" in binary_test_results[metric]
                    and binary_test_results[metric]["pvalue"] is not None
                ):
                    if binary_test_results[metric]["pvalue"] < alpha:
                        binary_test_results[metric]["Decision"] = "Reject H0"
                    else:
                        binary_test_results[metric]["Decision"] = "Accept H0"

        if self.continuous_metrics is not None:
            for metric in self.continuous_metrics:
                if (
                    "effect_estimate" in continuous_test_results[metric]
                    and continuous_test_results[metric]["effect_estimate"] is not None
                ):
                    continuous_test_results[metric][
                        "confidence_interval_bootstrapped"
                    ] = ConfidenceIntervals(
                        df=self.df,
                        user_id_column=self.user_id_column,
                        treatment=self.treatment,
                        metric=metric,
                        metric_type="Continuous",
                        method=method_continuous,
                        action_date=self.action_date,
                    ).bootstrap_confidence_interval(
                        alpha=alpha
                    )
                if (
                    "pvalue" in continuous_test_results[metric]
                    and continuous_test_results[metric]["pvalue"] is not None
                ):
                    if continuous_test_results[metric]["pvalue"] < alpha:
                        continuous_test_results[metric]["Decision"] = "Reject H0"
                    else:
                        continuous_test_results[metric]["Decision"] = "Accept H0"

        return {**binary_test_results, **continuous_test_results}


def multiple_hypothesis_correction(pvals, method="bonferroni"):
    correction = sms.multitest.multipletests(pvals=pvals, method=method)
    return correction



import streamlit as st
import numpy as np
import plotly.graph_objects as go

from tw_experimentation.setuper import (
    Setuper,
    effect_size_to_uplift,
)
from tw_experimentation.streamlit.streamlit_utils import (
    initalise_session_states,
    wrapper_auto_calc_pre_experiment,
    wrap_button_auto_calc_pre_experiment,
)


def page_2_experiment_design():
    st.session_state.update(st.session_state)

    EXPERIMENT_PAGE = "Experiment Design"
    initalise_session_states()

    st.session_state["last_page"] = EXPERIMENT_PAGE

    st.header("Sample Size Calculation")

    st.markdown(
        r"""
        We first determine the targeted sample size based on the minimal detectable effect.

        If we want to document a certain uplift, we also need to provide
        - an estimate of the population mean (e.g., baseline conversion for binary outcomes),
        - treatment share,
        - type-I error $\alpha$, and
        - type-II error $\beta$. (***$\beta$ = 1 - Power***)
        - effect relation to base metric / mean: 'absolute' or 'relative'


        """
    )
    st.write(
        "The sample size calculation is based on a two-sample proportion z-test for binary"
        " metrics and"
        + "two-sample t-test with equal variance for continuous outcome metrics."
    )

    targets_pre_experiment = (
        st.session_state.ed.pre_experiment_cols
        if st.session_state.ed is not None
        else None
    )
    if targets_pre_experiment is not None and len(targets_pre_experiment) > 0:
        _ = wrapper_auto_calc_pre_experiment(st.session_state.ed)
        st.divider()
        with st.expander("Auto-calculate from pre-experiment data"):
            st.markdown(
                r"""
                If you have pre-experiment data, you can use it to estimate the targeted sample size.
                The method will auto-fill
                - the target metric type,
                - the baseline value, and
                - the standard deviation (for continuous metrics).
                """
            )
            st.selectbox(
                "Target metric",
                st.session_state["pre_experiment"],
                key="exp_design_target",
            )

            if st.session_state["exp_design_target"] is not None:
                st.button(
                    "Auto_calculate",
                    on_click=wrap_button_auto_calc_pre_experiment,
                    args=(st.session_state["exp_design_target"],),
                )

    st.divider()

    col11, col12 = st.columns(2)

    with col11:
        st.number_input(
            r"$\alpha$ (Type-I error, e.g., 1.0 = 1%)",
            min_value=0.0,
            max_value=100.0,
            step=0.01,
            key="exp_design_alpha",
            format="%.2f",
        )

    with col12:
        st.number_input(
            r"$\beta$ (Type-II error, e.g., 1.0 = 1%)",
            min_value=0.0,
            max_value=100.0,
            step=0.01,
            key="exp_design_beta",
            format="%.2f",
        )

    st.number_input(
        "Treatment share (e.g., 50.0 = 50%)",
        min_value=0.0,
        max_value=100.0,
        step=0.01,
        key="exp_design_treatment_share",
        format="%.2f",
    )

    EFFECT_TYPES = ["relative", "absolute"]

    col21, col22 = st.columns(2)

    with col21:
        st.selectbox("Effect type", EFFECT_TYPES, key="exp_design_effect_type")

    with col22:
        mde_max_val = None
        relative_effect = st.session_state["exp_design_effect_type"] == "relative"
        if relative_effect:
            mde_max_val = 1.0

        st.number_input(
            (
                "Uplift to detect"
                if not relative_effect
                else "Relative uplift to detect (e.g., 0.1 = 10%)"
            ),
            min_value=0.001,
            max_value=mde_max_val,
            step=0.001,
            key="exp_design_mde",
            format="%.3f",
        )

    METRIC_TYPE = ["binary", "continuous"]
    st.selectbox("Key outcome metric type", METRIC_TYPE, key="exp_design_metric_type")

    if st.session_state["exp_design_metric_type"] == "binary":
        st.number_input(
            "Baseline value",
            min_value=0.0,
            max_value=1.0,
            step=0.001,
            key="exp_design_baseline_conversion",
            format="%.3f",
        )
    elif st.session_state["exp_design_metric_type"] == "continuous":
        st.number_input("Baseline mean", min_value=0.0, key="exp_design_baseline_mean")
        st.number_input("Standard deviation", min_value=0.0, key="exp_design_sd")

    # Plotting

    if st.session_state["exp_design_metric_type"] == "binary":
        sd = np.sqrt(
            st.session_state["exp_design_baseline_conversion"]
            * (1 - st.session_state["exp_design_baseline_conversion"])
        )
        setup = Setuper.from_uplift(
            alpha=st.session_state["exp_design_alpha"] / 100,
            beta=st.session_state["exp_design_beta"] / 100,
            uplift=st.session_state["exp_design_mde"],
            sd=sd,
            mean=st.session_state["exp_design_baseline_conversion"],
            relation=st.session_state["exp_design_effect_type"],
            treatment_share=st.session_state["exp_design_treatment_share"] / 100,
        )
        sample_size = setup.sample_size_two_sample_proportion_z_test()
        max_sample_size = 2 * sample_size["Total Sample Size"]
        x_sample_size = np.linspace(
            200 if max_sample_size > 500 else 20, max_sample_size, num=100
        )

        def uplift_map(x):
            effect_size_to_uplift(
                setup.effect_size_two_sample_z_test(x),
                st.session_state["exp_design_baseline_conversion"],
                sd,
                relation=st.session_state["exp_design_effect_type"],
            )

        uplift = np.array(list(map(uplift_map, x_sample_size)))

        fig = go.FigureWidget(data=go.Scatter(x=x_sample_size, y=uplift))
        fig.add_vline(
            x=sample_size["Total Sample Size"],
            line_width=3,
            line_dash="dash",
            line_color="green",
        )
        fig.update_xaxes(title_text="Sample Size")
        fig.update_yaxes(title_text="Minimum Detectable Uplift")

    elif st.session_state["exp_design_metric_type"] == "continuous":
        setup = Setuper.from_uplift(
            alpha=st.session_state["exp_design_alpha"] / 100,
            beta=st.session_state["exp_design_beta"] / 100,
            uplift=st.session_state["exp_design_mde"],
            sd=st.session_state["exp_design_sd"],
            mean=st.session_state["exp_design_baseline_mean"],
            relation=st.session_state["exp_design_effect_type"],
            treatment_share=st.session_state["exp_design_treatment_share"] / 100,
        )
        sample_size = setup.sample_size_t_test()
        max_sample_size = 2 * sample_size["Total Sample Size"]
        x_sample_size = np.linspace(
            200 if max_sample_size > 500 else 20, max_sample_size, num=100
        )

        def uplift_map(x):
            effect_size_to_uplift(
                setup.effect_size_t_test(x),
                st.session_state["exp_design_baseline_mean"],
                st.session_state["exp_design_sd"],
                relation=st.session_state["exp_design_effect_type"],
            )

        uplift = np.array(list(map(uplift_map, x_sample_size)))
        fig = go.FigureWidget(data=go.Scatter(x=x_sample_size, y=uplift))
        fig.add_vline(
            x=sample_size["Total Sample Size"],
            line_width=3,
            line_dash="dash",
            line_color="green",
        )
        fig.update_xaxes(title_text="Sample Size")
        fig.update_yaxes(title_text="Minimum Detectable Uplift")

    col_ss_1, col_ss_2, col_ss_3 = st.columns(3)

    with col_ss_1:
        st.metric(
            "Total Sample Size",
            value=sample_size["Treatment Sample Size"]
            + sample_size["Control Sample Size"],
        )

    with col_ss_2:
        st.metric(
            "Treatment Sample Size",
            value=sample_size["Treatment Sample Size"],
        )

    with col_ss_3:
        st.metric(
            "Control Sample Size",
            value=sample_size["Control Sample Size"],
        )

    st.plotly_chart(fig)



from jax import lax
import jax.numpy as jnp
import jax.random as random

from numpyro.distributions import constraints
from numpyro.distributions.distribution import Distribution
from numpyro.distributions.util import (
    is_prng_key,
    lazy_property,
    promote_shapes,
    validate_sample,
)
from numpyro.distributions import LogNormal


class ZeroInflatedProbsPatch(Distribution):
    """ZeroInflatedProbs distribution from Numpyro.

    https://num.pyro.ai/en/stable/_modules/numpyro/distributions/discrete.html#ZeroInflatedDistribution # noqa: E501

    Remove assertion check for base_dist.support.is_discrete to allow for
        continuous base_dist
    i.e. ZeroInflatedLogNormal
    """

    arg_constraints = {"gate": constraints.unit_interval}

    def __init__(self, base_dist, gate, *, validate_args=None):
        batch_shape = lax.broadcast_shapes(jnp.shape(gate), base_dist.batch_shape)
        (self.gate,) = promote_shapes(gate, shape=batch_shape)
        # assert base_dist.support.is_discrete
        if base_dist.event_shape:
            warning = "ZeroInflatedProbs expected empty base_dist.event_shape but got"
            raise ValueError(f"{warning} {base_dist.event_shape}")
        # XXX: we might need to promote parameters of base_dist but let's keep
        # this simplified for now
        self.base_dist = base_dist.expand(batch_shape)
        super(ZeroInflatedProbsPatch, self).__init__(
            batch_shape, validate_args=validate_args
        )

    def sample(self, key, sample_shape=()):
        assert is_prng_key(key)
        key_bern, key_base = random.split(key)
        shape = sample_shape + self.batch_shape
        mask = random.bernoulli(key_bern, self.gate, shape)
        samples = self.base_dist(rng_key=key_base, sample_shape=sample_shape)
        return jnp.where(mask, 0, samples)

    @validate_sample
    def log_prob(self, value):
        log_prob = jnp.log1p(-self.gate) + self.base_dist.log_prob(value)
        return jnp.where(value == 0, jnp.log(self.gate + jnp.exp(log_prob)), log_prob)

    @constraints.dependent_property(is_discrete=True, event_dim=0)
    def support(self):
        return self.base_dist.support

    @lazy_property
    def mean(self):
        return (1 - self.gate) * self.base_dist.mean

    @lazy_property
    def variance(self):
        return (1 - self.gate) * (
            self.base_dist.mean**2 + self.base_dist.variance
        ) - self.mean**2

    @property
    def has_enumerate_support(self):
        return self.base_dist.has_enumerate_support

    def enumerate_support(self, expand=True):
        return self.base_dist.enumerate_support(expand=expand)


class ZeroInflatedLogNormal(ZeroInflatedProbsPatch):
    arg_constraints = {
        "gate": constraints.unit_interval,
        "scale": constraints.positive,
        "loc": constraints.real,
    }
    support = constraints.real

    def __init__(self, gate=0.5, loc=0.0, scale=1.0, *, validate_args=None):
        _, self.loc, self.scale = promote_shapes(gate, loc, scale)
        super().__init__(
            LogNormal(loc=self.loc, scale=self.scale), gate, validate_args=validate_args
        )
